import numpy as np

from .. import errors
from ..segment import Segment
from ..filesystem import VTOC, Dirent, Directory, Filesystem
from ..file_type import guess_file_type

import logging
log = logging.getLogger(__name__)
try:  # Expensive debugging
    _xd = _expensive_debugging
except NameError:
    _xd = False


class Dos33BootSegment(Segment):
    ui_name = "DOS 3.3 Boot"
    def __init__(self, filesystem):
        media = filesystem.media
        Segment.__init__(self, media, 0, 0, name="Boot Sectors", length=256*37)
        self.segments = self.calc_boot_segments()

    def calc_boot_segments(self):
        boot1 = Segment(self, 0, 0x800, "Boot 1", length=256)
        boot2 = Segment(self, 256, 0x3700, "Boot 2", length=9*256)
        relocator = Segment(self, 10*256, 0x1b00, "Relocator", length=2*256)
        boot3 = Segment(self, 12*256, 0x1d00, "Boot 3", length=25*256)
        return [boot1, boot2, relocator, boot3]


class Dos33VTOC(VTOC):
    ui_name = "DOS 3.3 VTOC"
    max_tracks = (256 - 0x38) // 4  # 50, but kept here in case sector size changed
    max_sector = max_tracks * 16
    vtoc_bit_reorder_index = np.tile(np.arange(15, -1, -1), max_tracks) + (np.repeat(np.arange(max_tracks), 16) * 16)

    vtoc_type = np.dtype([
        ('unused1', 'S1'),
        ('cat_track','u1'),
        ('cat_sector','u1'),
        ('dos_release', 'u1'),
        ('unused2', 'S2'),
        ('vol_num', 'u1'),
        ('unused3', 'S32'),
        ('max_pairs', 'u1'),
        ('unused4', 'S8'),
        ('last_track', 'u1'),
        ('track_dir', 'i1'),
        ('unused5', 'S2'),
        ('num_tracks', 'u1'),
        ('sectors_per_track', 'u1'),
        ('sector_size', 'u2'),
        ])

    def find_segment_location(self):
        media = self.media
        index = 17 * 16
        if not media.is_sector_valid(index):
            raise errors.FilesystemError(f"Media ends before track 17")
        return media.get_contiguous_sectors_offsets(index, 1)

    def calc_sector_map_size(self):
        values = self[0:self.vtoc_type.itemsize].view(dtype=self.vtoc_type)[0]
        print(self.container_offset)
        print(values)
        media = self.media
        media.first_directory_track = int(values['cat_track'])
        media.first_directory_sector = int(values['cat_sector'])
        media.sector_size = int(values['sector_size'])
        media.max_sectors = int(values['num_tracks']) * int(values['sectors_per_track'])
        media.ts_pairs = int(values['max_pairs'])
        media.dos_release = int(values['dos_release'])
        media.last_track_num = int(values['last_track'])
        media.track_alloc_dir = int(values['track_dir'])
        return self.max_sector

    def unpack_vtoc(self):
        # VTOC stored in groups of 4 bytes starting at 0x38
        # in bits, the sector used data is stored by track:
        #
        # FEDCBA98 76543210 xxxxxxxx xxxxxxxx
        #
        # where the x values are ignored (should be zeros). Track 0 info is
        # found starting at 0x38, track 1 is found at 0x3c, etc.
        #
        # Want to convert this to an array that is a list of bits by
        # track/sector number, i.e.:
        #
        # t0s0 t0s1 t0s2 t0s3 t0s4 t0s5 t0s6 t0s7 ... t1s0 t1s1 ... etc
        #
        # Problem: the bits are stored backwards, so a straight unpackbits will
        # produce:
        #
        # t0sf t0se t0sd ...
        #
        # i.e. each group of 16 bits needs to be reversed.

        # create a view starting at 0x38 where out of every 4 bytes, the first
        # two are used and the second 2 are skipped. Regular slicing doesn't
        # work like this, so thanks to stackoverflow.com/questions/33801170,
        # reshaping it to a 2d array with 4 elements in each row, doing a slice
        # *there* to skip the last 2 entries in each row, then flattening it
        # gives us what we need.
        usedbytes = self[0x38:].reshape((-1, 4))[:,:2].flatten()

        # The bits here are still ordered backwards for each track, e.g. F E D
        # C B A 9 8 7 6 5 4 3 2 1 0
        bits = np.unpackbits(usedbytes)

        # so we need to reorder them using numpy's indexing before stuffing
        # them into the sector map
        self.sector_map[0:self.max_sector] = bits[self.vtoc_bit_reorder_index]
        if _xd: log.debug("vtoc before:\n%s" % str(self))  # expensive debugging call

    def pack_vtoc(self):
        if _xd: log.debug("vtoc after:\n%s" % str(self))  # expensive debugging call

        # reverse the process from above, so swap the order of every 16 bits,
        # turn them into bytes, then stuff them back into the vtoc. The bit
        # reorder list is commutative, so we don't need another order here.
        packed = np.packbits(self.sector_map[self.vtoc_bit_reorder_index])
        vtoc = self[0x38:].reshape((-1, 4))
        packed = packed.reshape((-1, 2))
        vtoc[:,:2] = packed[:,:]

        # FIXME
        self[0x38:] = vtoc.flatten()


class Dos33Dirent(Dirent):
    ui_name = "DOS 3.3 Dirent"
    format = np.dtype([
        ('track', 'u1'),
        ('sector', 'u1'),
        ('flag', 'u1'),
        ('name','S30'),
        ('num_sectors','<u2'),
        ])

    def __init__(self, directory, filenum):
        start = self.format.itemsize * filenum
        Dirent.__init__(self, directory, filenum, start, self.format.itemsize)
        self._file_type = 0
        self.locked = False
        self.deleted = False
        self.track = 0
        self.sector = 0
        self.filename = ""
        self.num_sectors = 0
        self.current_sector_index = 0
        self.current_read = 0
        self.sectors_seen = None
        self.sector_map = None

    def __str__(self):
        return "File #%-2d (%s) %03d %-30s %03d %03d" % (self.file_num, self.verbose_info, self.num_sectors, self.filename, self.track, self.sector)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.filename == other.filename and self.track == other.track and self.sector == other.sector and self.num_sectors == other.num_sectors

    type_to_text = {
        0x0: "T",  # text
        0x1: "I",  # integer basic
        0x2: "A",  # applesoft basic
        0x4: "B",  # binary
        0x8: "S",  # ?
        0x10: "R",  # relocatable object module
        0x20: "a",  # ?
        0x40: "b",  # ?
    }
    text_to_type = {v: k for k, v in type_to_text.items()}

    @property
    def file_type(self):
        """User friendly version of file type, not the binary number"""
        return self.type_to_text.get(self._file_type, "?")

    @property
    def verbose_info(self):
        if self.deleted:
            locked = "D"
            file_type = " "
        else:
            locked = "*" if self.locked else " "
            file_type = self.file_type
        flag = "%s%s" % (locked, file_type)
        return flag

    @property
    def flag(self):
        return 0xff if self.deleted else self._file_type | (0x80 * int(self.locked))

    def parse_raw_dirent(self):
        data = self.data[0:self.format.itemsize]
        values = data.view(dtype=self.format)[0]
        self.track = values[0]
        if self.track == 0xff:
            self.deleted = True
            self.track = data[0x20]
        else:
            self.deleted = False
        self.sector = values[1]
        self._file_type = values[2] & 0x7f
        self.locked = values[2] & 0x80
        self.filename = (data[3:0x20] - 0x80).tobytes().rstrip().decode("ascii", errors='ignore')
        self.num_sectors = int(values[4])
        self.is_sane = self.sanity_check()

    def encode_dirent(self):
        data = np.zeros([self.format.itemsize], dtype=np.uint8)
        values = data.view(dtype=self.format)[0]
        values[0] = 0xff if self.deleted else self.track
        values[1] = self.sector
        values[2] = self.flag
        n = min(len(self.filename), 30)
        data[3:3+n] = np.frombuffer(self.filename.encode("ascii"), dtype=np.uint8) | 0x80
        data[3+n:] = ord(' ') | 0x80
        if self.deleted:
            data[0x20] = self.track
        values[4] = self.num_sectors
        return data

    def sanity_check(self):
        media = self.filesystem.media
        if self.deleted:
            return True
        if self.track == 0:
            return False
        s = media.sector_from_track(self.track, self.sector)
        if not media.is_sector_valid(s):
            return False
        if self.num_sectors < 0 or self.num_sectors > media.max_sectors:
            return False
        return True

    def get_file(self):
        media = self.filesystem.media
        tslist = self.get_track_sector_list()
        offsets = media.follow_track_sector_list(tslist)
        if len(offsets) > 0:
            file_segment = guess_file_type(media, self.filename, offsets)
            self.segments = [tslist, file_segment]
            return file_segment

    def get_track_sector_list(self):
        media = self.filesystem.media
        offsets = media.follow_track_sector_pointers(self.track, self.sector, 0xc, 2 * 122)
        return Segment(media, offsets, 0, "Track/Sector List")

    def get_sectors_in_vtoc(self, image):
        self.get_track_sector_list(image)
        sectors = BaseSectorList(media)
        sectors.extend(self.track_sector_list)
        for sector_num in self.sector_map:
            sector = WriteableSector(media.sector_size, None, sector_num)
            sectors.append(sector)
        return sectors

    def get_filename(self):
        return self.filename

    def set_values(self, filename, filetype, index):
        self.filename = '%-30s' % filename[0:30]
        self._file_type = self.text_to_type.get(filetype, 0x04)
        self.locked = False
        self.deleted = False

    def get_binary_start_address(self, image):
        self.start_read(image)
        data, _, _, _ = self.read_sector(image)
        addr = int(data[0]) + 256 * int(data[1])
        return addr


class Dos33Directory(Directory):
    ui_name = "DOS 3.3 Directory"

    def find_segment_location(self):
        media = self.media
        offsets = media.follow_track_sector_pointers(media.first_directory_track, media.first_directory_sector, 0x0b, 7 * 0x23)
        return offsets, 0

    def calc_dirents(self):
        segments = []
        index = 0
        filenum = 0
        while index < len(self):
            dirent = Dos33Dirent(self, filenum)
            if dirent.in_use:
                dirent.set_comment_at(0x00, "FILE #%d: Track number of next catalog sector" % filenum)
                dirent.set_comment_at(0x01, "FILE #%d: Sector number of next catalog sector" % filenum)
                dirent.set_comment_at(0x02, "FILE #%d: File type" % filenum)
                dirent.set_comment_at(0x03, "FILE #%d: Filename" % filenum)
                dirent.set_comment_at(0x21, "FILE #%d: Number of sectors in file" % filenum)
                segments.append(dirent)
            index += len(dirent)
            filenum += 1
        return segments


class AppleDos33(Filesystem):
    ui_name = "Apple DOS 3.3"
    default_executable_extension = "BIN"

    def check_media(self, media):
        try:
            media.sector_from_track
        except AttributeError:
            raise errors.IncompatibleMediaError("Apple DOS needs track/sector access")

    def calc_boot_segment(self):
        return Dos33BootSegment(self)

    def calc_vtoc_segment(self):
        return Dos33VTOC(self)

    def calc_directory_segment(self):
        return Dos33Directory(self)



import glob

import numpy as np

from mock import *

from atrip.container import guess_container
from atrip import errors


read_only_compressors = set(['.Z'])


class BaseContainerTest:
    base_path = None
    expected_mime = ""

    @pytest.mark.parametrize(("ext", "mod_name"), [
        ('', 'none'),
        ('.gz', 'gzip'),
        ('.bz2', 'bzip2'),
        ('.xz', 'lzma'),
        ('.lz4', 'lz4'),
        ('.Z', 'unix compress'),
        ('.dcm', 'dcm'),
    ])
    def test_container(self, ext, mod_name):
        pathname = os.path.abspath(os.path.join(os.path.dirname(__file__), self.base_path + ext))
        try:
            sample_data = np.fromfile(pathname, dtype=np.uint8)
        except OSError as e:
            pytest.skip(f"no source file {pathname}")
        else:
            container = guess_container(sample_data)
            assert container.decompression_order[0].compression_algorithm == mod_name

            output = "tmp." + os.path.basename(pathname)
            pathname = os.path.join(os.path.dirname(__file__), output)

            try:
                compressed = container.calc_packed_bytes()
            except errors.InvalidAlgorithm:
                if ext in read_only_compressors:
                    pytest.skip(f"skipping {pathname} because can't compress {ext} yet")

            # compressed data may not be the same; don't really care as long as
            # it decompresses the same
            container2 = guess_container(compressed)
            assert np.array_equal(container._data, container2._data)


class TestContainerAtariDosSDImage(BaseContainerTest):
    base_path = "../samples/dos_sd_test1.atr"
    expected_mime = "application/vnd.atari8bit.atr"
    num_files_in_sample = 5

class TestContainerAtariDosEDImage(BaseContainerTest):
    base_path = "../samples/dos_ed_test1.atr"
    expected_mime = "application/vnd.atari8bit.atr"
    num_files_in_sample = 5

class TestContainerAtariDosDDImage(BaseContainerTest):
    base_path = "../samples/dos_dd_test1.atr"
    expected_mime = "application/vnd.atari8bit.atr"
    num_files_in_sample = 5


class TestMultipleCompression:
    @pytest.mark.parametrize(("pathname"), globbed_sample_atari_files)
    def test_glob(self, pathname):
        sample_data = np.fromfile(pathname, dtype=np.uint8)
        container = guess_container(sample_data)
        container.guess_media_type()
        output = "tmp." + os.path.basename(pathname)
        pathname = os.path.join(os.path.dirname(__file__), output)

        try:
            compressed = container.calc_packed_bytes()
        except errors.InvalidAlgorithm as e:
            pytest.skip(f"skipping {pathname}: {e}")
        else:
            # compressed data may not be the same; don't really care as long as
            # it decompresses the same
            container2 = guess_container(compressed)
            assert np.array_equal(container._data, container2._data)


if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.DEBUG)
    c = TestContainerAtariDosSDImage()
    # c.test_container(".gz", "gzip")
    # c.test_container(".dcm", "dcm")
    c = TestMultipleCompression()
    c.test_glob("../samples/mydos_sd_mydos4534.dcm.lz4")



import numpy as np

import logging
log = logging.getLogger(__name__)


magic = [
    {'mime': "application/x.atari8bit.atr;getaway_pd",
     'name': "Getaway Public Domain ATR",
     'signature': [
         (slice(8, 10), [0x82, 0x39]),
         (slice(12, 16), [0x67, 0x21, 0x70, 0x64]),
         ],
    },

    {'mime': "application/x.atari8bit.xex;getaway",
     'name': "Getaway XEX",
     'signature': [
         (slice(0, 6), [0xff, 0xff, 0x80, 0x2a, 0xff, 0x8a]),
         ],
    },

    {'mime': "application/x.atari8bit.atr.getaway",
     'name': "Getaway ATR",
     'signature': [
         (slice(0x10, 0x19), [0x00, 0xc1, 0x80, 0x0f, 0xcc, 0x22, 0x18, 0x60, 0x0e]),
         ],
    },

    {'mime': "application/x.atari8bit.atr;jumpman_level_tester",
     'name': "Jumpman Level Tester from Omnivore",
     'signature': [
         (slice(0, 5), [0x96, 0x02 , 0xd0 , 0x05 , 0x80]),
         (0x0196 + 0x3f, 0x4c),
         (0x0196 + 0x48, 0x20),
         (0x0196 + 0x4b, 0x60),
         (0x0196 + 0x4c, 0xff),
         ],
    },

    {'mime': "application/x.atari8bit.atr;jumpman",
     'name': "Jumpman",
     'signature': [
         (slice(0, 5), [0x96, 0x02 , 0x80 , 0x16 , 0x80]),
         (0x0810 + 0x3f, 0x4c),
         (0x0810 + 0x48, 0x20),
         (0x0810 + 0x4b, 0x60),
         (0x0810 + 0x4c, 0xff),
         ],
    },
]


def check_signature(raw, sig):
    for index, expected in sig:
        actual = raw.data[index].tolist()
        if actual == expected:
            log.debug(" match at %s: %s" % (str(index), str(expected)))
        if actual != expected:
            log.debug(" failed at %s: %s != %s" % (str(index), str(expected), str(raw.data[index])))
            return False
    return True


def guess_detail_for_mime(mime, raw, parser):
    for entry in magic:
        if entry['mime'].startswith(mime):
            log.debug("checking entry for %s" % entry['mime'])
            if check_signature(raw, entry['signature']):
                log.debug("found match: %s" % entry['name'])
                return entry['mime']
    return mime




import os
import sys

import wx

from sawx.utils.nputil import intscale
from sawx.ui import compactgrid as cg

from ..ui.segment_grid import SegmentGridControl, SegmentTable
from ..arch.bitmap_renderers import valid_bitmap_renderers

from .antic import AnticColorViewer

import logging
log = logging.getLogger(__name__)


class BitmapImageCache(cg.DrawTextImageCache):
    def draw_item(self, grid_control, dc, rect, data, style):
        start = 0
        end = len(data)
        nr = 1
        array = grid_control.bitmap_renderer.get_image(grid_control.segment_viewer, end, nr, end, data, style)
        width = array.shape[1]
        height = array.shape[0]
        if width > 0 and height > 0:
            array = intscale(array, grid_control.zoom_h * grid_control.scale_height, grid_control.zoom_w * grid_control.scale_width)
            #print("bitmap: %d,%d,3 after scaling: %s" % (height, width, str(array.shape)))
            image = wx.Image(array.shape[1], array.shape[0])
            image.SetData(array.tobytes())
            bmp = wx.Bitmap(image)
            dc.DrawBitmap(bmp, rect.x, rect.y)


class BitmapLineRenderer(cg.TableLineRenderer):
    default_image_cache = BitmapImageCache

    def __init__(self, grid_control, image_cache=None):
        image_cache = BitmapImageCache()
        w, h = self.calc_cell_size_in_pixels(grid_control)
        cg.LineRenderer.__init__(self, grid_control, w, h, grid_control.items_per_row, image_cache)

    def calc_cell_size_in_pixels(self, grid_control):
        w = grid_control.zoom_w * grid_control.scale_width * grid_control.pixels_per_byte
        h = grid_control.zoom_h * grid_control.scale_height
        return w, h

    # BaseLineRenderer interface

    def draw_line(self, grid_control, dc, line_num, col, index, last_index):
        t = grid_control.table
        rect = self.col_to_rect(line_num, col)
        data = t.data[index:last_index]
        style = t.style[index:last_index]
        self.image_cache.draw_item(grid_control, dc, rect, data, style)

    # fast BaseLineRenderer interface drawing entire grid at once

    def calc_bytes_per_row(self, table):
        return table.items_per_row

    def draw_grid(self, grid_control, dc, first_row, visible_rows, first_cell, visible_cells):
        t = grid_control.table
        log.debug(f"draw_grid: first_row={first_row} visible_rows={visible_rows}; first_cell={first_cell} visible_cells={visible_cells}")
        first_col = self.cell_to_col(first_row, first_cell)
        last_cell = min(first_cell + visible_cells, self.num_cells)
        last_row = min(first_row + visible_rows, t.num_rows)
        last_col = self.cell_to_col(last_row, last_cell - 1) + 1
        log.debug("draw_grid: rows:%d,%d (vis %d, num %d) cols:%d,%d" % (first_row, last_row, visible_rows, t.num_rows, first_col, last_col))

        ul_rect = self.col_to_rect(first_row, first_col)
        lr_rect = self.col_to_rect(last_row - 1, last_col - 1)
        frame_rect = wx.Rect(ul_rect.x, ul_rect.y, lr_rect.x - ul_rect.x + lr_rect.width, lr_rect.y - ul_rect.y + lr_rect.height)
        dc.SetClippingRegion(frame_rect)

        # First and last rows may not span entire width. Process those
        # separately
        #
        # First row may not have bytes at the beginning of the row if the start
        # offset is not zero
        if first_row == 0:
            try:
                col, index, last_index = self.calc_column_range(grid_control, first_row, first_col, last_col)
            except IndexError:
                pass  # skip lines with no visible cells
            else:
                self.draw_line(grid_control, dc, first_row, col, index, last_index)
            first_row += 1
            if first_row == last_row:
                return
            frame_rect.y += ul_rect.height

        # Last row may not have bytes at the end of the row
        if last_row == t.num_rows:
            try:
                col, index, last_index = self.calc_column_range(grid_control, last_row - 1, first_col, last_col)
            except IndexError:
                pass  # skip lines with no visible cells
            else:
                self.draw_line(grid_control, dc, last_row - 1, col, index, last_index)
            last_row -= 1

        # If there are any more rows to display, they will be full-width rows;
        # i.e. the data is in a rectangular grid
        nr = last_row - first_row
        if nr > 0:
            bytes_per_row = self.calc_bytes_per_row(t)
            nc = last_col - first_col
            offset = t.start_offset % bytes_per_row
            first_index = (first_row * bytes_per_row) - offset
            last_index = (last_row * bytes_per_row) - offset
            log.debug(f"drawing rectangular grid: bpr={bytes_per_row}, first,last={first_index},{last_index}, nr={nr}, start_offset={offset}")
            data = t.data[first_index:last_index].reshape((nr, bytes_per_row))[0:nr,first_col:last_col].flatten()
            style = t.style[first_index:last_index].reshape((nr, bytes_per_row))[0:nr,first_col:last_col].flatten()

            # get_image(cls, machine, antic_font, byte_values, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols):

            array = grid_control.bitmap_renderer.get_image(grid_control.segment_viewer, nc, nr, nc * nr, data, style)
            width = array.shape[1]
            height = array.shape[0]
            if width > 0 and height > 0:
                array = intscale(array, grid_control.zoom_h, grid_control.zoom_w)
                #print("bitmap: %d,%d,3 after scaling: %s" % (height, width, str(array.shape)))
                image = wx.Image(array.shape[1], array.shape[0])
                image.SetData(array.tobytes())
                bmp = wx.Bitmap(image)
                dc.DrawBitmap(bmp, frame_rect.x, frame_rect.y)


class BitmapGridControl(SegmentGridControl):
    def set_viewer_defaults(self):
        self.items_per_row = self.view_params.bitmap_width
        self.zoom = 2

    @property
    def bitmap_renderer(self):
        return self.segment_viewer.bitmap_renderer

    @property
    def zoom_w(self):
        return self.zoom  # * self.bitmap_renderer.scale_width

    @property
    def zoom_h(self):
        return self.zoom  # * self.bitmap_renderer.scale_height

    @property
    def pixels_per_byte(self):
        return self.bitmap_renderer.pixels_per_byte

    @property
    def scale_width(self):
        return self.bitmap_renderer.scale_width

    @property
    def scale_height(self):
        return self.bitmap_renderer.scale_height

    def calc_line_renderer(self):
        if hasattr(self, 'segment_viewer'):
            return BitmapLineRenderer(self)
        return SegmentGridControl.calc_line_renderer(self)

    def verify_line_renderer(self):
        self.recalc_line_renderer()


class BitmapViewer(AnticColorViewer):
    name = "bitmap"

    ui_name = "Bitmap"

    control_cls = BitmapGridControl

    has_bitmap = True

    has_width = True

    width_text = "bitmap width in bytes"

    has_zoom = True

    zoom_text = "bitmap zoom factor"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bitmap_renderer_name = "B/W, 1bpp, on=black"
        self._bitmap_renderer = None

    @property
    def bitmap_renderer_name(self):
        return self._bitmap_renderer_name

    @bitmap_renderer_name.setter
    def bitmap_renderer_name(self, value):
        self._bitmap_renderer_name = value
        self._bitmap_renderer = None
        self.graphics_properties_changed()

    @property
    def bitmap_renderer(self):
        if self._bitmap_renderer is None:
            self._bitmap_renderer = valid_bitmap_renderers[self._bitmap_renderer_name]
        return self._bitmap_renderer

    @property
    def window_title(self):
        return self.bitmap_renderer.name

    def validate_width(self, width):
        return self.bitmap_renderer.validate_bytes_per_row(width)

    def recalc_data_model(self):
        self.control.recalc_view()


class MemoryMapViewer(BitmapViewer):
    name = "memmap"

    ui_name = "Memory Page Map"



"""Baseclass for actions with common code applicable to viewers
"""
from sawx.action import SawxAction, SawxSubAction, SawxListAction, SawxRadioAction, SawxRadioListAction

import logging
log = logging.getLogger(__name__)


class ViewerActionMixin:
    @property
    def viewer(self):
        try:
            return self.popup_data["popup_viewer"]
        except (KeyError, TypeError) as e:
            return self.editor.focused_viewer

    @property
    def linked_base(self):
        return self.editor.focused_viewer.linked_base


class ViewerAction(ViewerActionMixin, SawxAction):
    pass


class ViewerSubAction(ViewerActionMixin, SawxSubAction):
    pass


class ViewerRadioAction(ViewerActionMixin, SawxRadioAction):
    pass


class ViewerListAction(ViewerActionMixin, SawxListAction):
    pass


class ViewerRadioListAction(ViewerActionMixin, SawxRadioListAction):
    pass



class select_hgr(ViewerAction):
    name = "Select HGR Without Screen Holes"

    def perform(self, action_key):
        self.viewer.select_hgr()
        self.editor.document.refresh_event(flags=True)



#!/usr/bin/env python
#
# Note: can't figure out a way to determine the offset of structure members
# after construction, so I've inserted additional entries into the structure
# that contain the raw offset into the file.
#
# I generated these using the regex search/replace (note the replace is multi-
# line):
#
# find: \"(\w+)\" */ *(.+)
# replace: "\1_offset" / Tell,
#          "\1\" / \2

import numpy as np

from construct import *

import logging
log = logging.getLogger(__name__)


Atari800_MACHINE_XLXE = 1
MEMORY_RAM_320_RAMBO = 320
MEMORY_RAM_320_COMPY_SHOP = 321
BB_RAM_SIZE = 0x10000
COMMAND_FRAME_SIZE = 6
DATA_BUFFER_SIZE = 256 + 3

def init_atari800_struct():
    Fname = Struct(
        "len_offset" / Tell,
        "len" / Int16ul,
        "name_offset" / Tell,
        "name" / String(this.len),
        )

    Sio = Struct(
        "status_offset" / Tell,
        "status" / Int32sl,
        "filename_offset" / Tell,
        "filename" / Fname,
        )

    a8save = "a8save" / Struct(
        "signature_segment" / Tell,
        "signature_offset" / Tell,
        "signature" / Struct(
            "magic_offset" / Tell,
            "magic" / Const(b"ATARI800"),
            "version_offset" / Tell,
            "version" / Int8ul,
            "verbose_offset" / Tell,
            "verbose" / Int8ul,
            "signature_offset_end" / Tell,
            ),
        "signature_segment_end" / Tell,

        "machine_type_segment" / Tell,
        "machine_type_offset" / Tell,
        "machine_type" / Struct(
            "tv_mode_offset" / Tell,
            "tv_mode" / Int8ul,
            "id_offset" / Tell,
            "id" / Int8ul,
            "machine_type_offset_end" / Tell,
            ),
        "machine_type_segment_end" / Tell,

        "xl_xe_segment" / Tell,
        "xl_xe_offset" / Tell,
        "xl_xe" / If(this.machine_type.id > 0,
            Struct(
                "builtin_basic_offset" / Tell,
                "builtin_basic" / Int8ul,
                "keyboard_leds_offset" / Tell,
                "keyboard_leds" / Int8ul,
                "f_keys_offset" / Tell,
                "f_keys" / Int8ul,
                "jumper_offset" / Tell,
                "jumper" / Int8ul,
                "builtin_game_offset" / Tell,
                "builtin_game" / Int8ul,
                "keyboard_detached_offset" / Tell,
                "keyboard_detached" / Int8ul,
                "xl_xe_offset_end" / Tell,
                )
            ),
        "xl_xe_segment_end" / Tell,

        # CARTRIDGE
        "cartridge_segment" / Tell,
        "cartridge_offset" / Tell,
        "cartridge" / Struct(
            "cartridge_flag_offset" / Tell,
            "cartridge_flag" / Int32sl,
            "left_cart_filename_offset" / Tell,
            "left_cart_filename" / If(this.cartridge_flag != 0, Fname),
            "left_cart_state_offset" / Tell,
            "left_cart_state" / If(this.cartridge_flag != 0, Int32sl),
            "right_cart_type_offset" / Tell,
            "right_cart_type" / If(this.cartridge_flag < 0, Int32sl),
            "right_cart_filename_offset" / Tell,
            "right_cart_filename" /  If(this.cartridge_flag < 0, Fname),
            "right_cart_state_offset" / Tell,
            "right_cart_state" / If(this.cartridge_flag < 0, Int32sl),
            "cartridge_offset_end" / Tell,
            ),
        "cartridge_segment_end" / Tell,

        # SIO
        # StateSav_SaveINT((int *) &SIO_drive_status[i], 1);
        # StateSav_SaveFNAME(SIO_filename[i]);
        "SIO_1_offset" / Tell,
        "SIO_1" / Sio,
        "SIO_2_offset" / Tell,
        "SIO_2" / Sio,
        "SIO_3_offset" / Tell,
        "SIO_3" / Sio,
        "SIO_4_offset" / Tell,
        "SIO_4" / Sio,
        "SIO_5_offset" / Tell,
        "SIO_5" / Sio,
        "SIO_6_offset" / Tell,
        "SIO_6" / Sio,
        "SIO_7_offset" / Tell,
        "SIO_7" / Sio,
        "SIO_8_offset" / Tell,
        "SIO_8" / Sio,

        # ANTIC
        "ANTIC_segment" / Tell,
        "ANTIC_offset" / Tell,
        "ANTIC" / Struct(
            "DMACTL_offset" / Tell,
            "DMACTL" / Int8ul,
            "CHACTL_offset" / Tell,
            "CHACTL" / Int8ul,
            "HSCROL_offset" / Tell,
            "HSCROL" / Int8ul,
            "VSCROL_offset" / Tell,
            "VSCROL" / Int8ul,
            "PMBASE_offset" / Tell,
            "PMBASE" / Int8ul,
            "CHBASE_offset" / Tell,
            "CHBASE" / Int8ul,
            "NMIEN_offset" / Tell,
            "NMIEN" / Int8ul,
            "NMIST_offset" / Tell,
            "NMIST" / Int8ul,
            "_ir_offset" / Tell,
            "_ir" / Int8ul,
            "_anticmode_offset" / Tell,
            "_anticmode" / Int8ul,
            "_dctr_offset" / Tell,
            "_dctr" / Int8ul,
            "_lastline_offset" / Tell,
            "_lastline" / Int8ul,
            "_need_dl_offset" / Tell,
            "_need_dl" / Int8ul,
            "_vscrol_off_offset" / Tell,
            "_vscrol_off" / Int8ul,

            "DLIST_offset" / Tell,
            "DLIST" / Int16ul,
            "_screenaddr_offset" / Tell,
            "_screenaddr" / Int16ul,

            "xpos_offset" / Tell,
            "xpos" / Int32sl,
            "_xpos_limit_offset" / Tell,
            "_xpos_limit" / Int32sl,
            "ypos_offset" / Tell,
            "ypos" / Int32sl,
            "ANTIC_offset_end" / Tell,
            ),
        "ANTIC_segment_end" / Tell,

        #CPU
        "CPU_segment" / Tell,
        "CPU_offset" / Tell,
        "CPU" / Struct(
            "A_offset" / Tell,
            "A" / Int8ul,
            "P_offset" / Tell,
            "P" / Int8ul,
            "S_offset" / Tell,
            "S" / Int8ul,
            "X_offset" / Tell,
            "X" / Int8ul,
            "Y_offset" / Tell,
            "Y" / Int8ul,
            "IRQ_offset" / Tell,
            "IRQ" / Int8ul,
            "CPU_offset_end" / Tell,
            ),
        "CPU_segment_end" / Tell,

        #Memory
        "AXLON_segment" / Tell,
        "AXLON_offset" / Tell,
        "AXLON" / If(this.machine_type.id == 0,
            Struct(
                "num_banks_offset" / Tell,
                "num_banks" / Int8ul,
                "curbank_offset" / Tell,
                "curbank" / If (this.num_banks > 0, Int32sl),
                "0f_mirror_offset" / Tell,
                "0f_mirror" / If (this.num_banks > 0, Int32sl),
                "axlon_ram_offset" / Tell,
                "axlon_ram" / If (this.num_banks > 0, Bytes(this.num_banks * 0x4000)),
            "AXLON_offset_end" / Tell,
                )
            ),
        "AXLON_segment_end" / Tell,

        "MOSAIC_segment" / Tell,
        "MOSAIC_offset" / Tell,
        "MOSAIC" / If(this.machine_type.id == 0,
            Struct(
                "num_banks_offset" / Tell,
                "num_banks" / Int8ul,
                "curbank_offset" / Tell,
                "curbank" / If (this.num_banks > 0, Int32sl),
                "mosaic_ram_offset" / Tell,
                "mosaic_ram" / If (this.num_banks > 0, Bytes(this.num_banks * 0x1000)),
            "MOSAIC_offset_end" / Tell,
                )
            ),
        "MOSAIC_segment_end" / Tell,

        "RAM_segment" / Tell,
        "RAM_offset" / Tell,
        "ram" / Struct(
            "base_ram_size_kb_offset" / Tell,
            "base_ram_size_kb" / Int32sl,
            "ram_segment" / Tell,
            "ram_offset" / Tell,
            "ram" / Bytes(65536),
            "ram_segment_end" / Tell,
            "ram_attrib_segment" / Tell,
            "attrib_offset" / Tell,
            "attrib" / Bytes(65536),
            "ram_attrib_segment_end" / Tell,
            "RAM_offset_end" / Tell,
            ),
        "RAM_segment_end" / Tell,

        "XLRAM_segment" / Tell,
        "XLRAM_offset" / Tell,
        "XLRAM" / If(this.machine_type.id == Atari800_MACHINE_XLXE,
            Struct(
                "basic_$A000_segment" / Tell,
                "basic_offset" / Tell,
                "basic" / If(this._.signature.verbose != 0, Bytes(8192)),
                "basic_$A000_segment_end" / Tell,
                "under_cartA0BF_$A000_segment" / Tell,
                "under_cartA0BF_offset" / Tell,
                "under_cartA0BF" / Bytes(8192),
                "under_cartA0BF_$A000_segment_end" / Tell,
                "os_offset_$C000_segment" / Tell,
                "os_offset" / Tell,
                "os" / If(this._.signature.verbose != 0, Bytes(16384)),
                "os_offset_$C000_segment_end" / Tell,
                "under_atarixl_os_offset_$C000_segment" / Tell,
                "under_atarixl_os_offset" / Tell,
                "under_atarixl_os" / Bytes(16384),
                "under_atarixl_os_offset_$C000_segment" / Tell,
                "xegame_offset_$E000_segment_end" / Tell,
                "xegame_offset" / Tell,
                "xegame" / If(this._.signature.verbose != 0, Bytes(0x2000)),
                "xegame_offset_$E000_segment_end" / Tell,
            "XLRAM_offset_end" / Tell,
                )
            ),
        "XLRAM_segment_end" / Tell,

        "num_16k_xe_banks_offset" / Tell,
        "num_16k_xe_banks" / Int32sl,

        "ram_size_kb_before_rambo_offset" / Tell,
        "ram_size_kb_before_rambo" / Computed(this.ram.base_ram_size_kb + (this.num_16k_xe_banks * 16)),

        # rambo_kb will be populated if RAMBO = 320 (0x140), COMPY_SHOP = 321 (0x141)
        "rambo_kb_offset" / Tell,
        "rambo_kb" / IfThenElse(this.ram_size_kb_before_rambo & 0xfffe == 0x140, Int32sl, Computed(0)), # value is MEMORY_ram_size - 320

        "ram_size_offset" / Tell,
        "ram_size" / Computed(this.ram_size_kb_before_rambo + this.rambo_kb),

        "PORTB_offset" / Tell,
        "PORTB" / Int8ul,

        "CartA0BF_enabled_offset" / Tell,
        "CartA0BF_enabled" / Int32sl,

        "atarixe_memory_offset" / Tell,
        "atarixe_memory" / If(this.ram_size > 64,
            Struct(
                "size_offset" / Tell,
                "size" / Computed((1 + (this.ram_size - 64) / 16) * 16384),
                "xeram_offset" / Tell,
                "xeram" / Bytes(this.size),
                "selftest_enabled_offset" / Tell,
                "selftest_enabled" / Computed(lambda ctx: (ctx._.PORTB & 0x81) == 0x01 and not ((ctx._.PORTB & 0x30) != 0x30 and ctx._.ram_size == MEMORY_RAM_320_COMPY_SHOP) and not ((ctx._.PORTB & 0x10) == 0 and ctx._.ram_size == 1088)),
                "antic_bank_under_selftest_offset" / Tell,
                "antic_bank_under_selftest" / If(this.selftest_enabled, Bytes(0x800)),
            "atarixe_memory_offset_end" / Tell,
                )
            ),

        "mapram_segment" / Tell,
        "mapram_offset" / Tell,
        "mapram" / If(lambda ctx: ctx.machine_type.id == Atari800_MACHINE_XLXE and ctx.ram_size > 20,
            Struct(
                "enable_offset" / Tell,
                "enable" / Int32sl,
                "memory_offset" / Tell,
                "memory" / If(this.enable != 0, Bytes(0x800)),
                "mapram_offset_end" / Tell,
                )
            ),
        "mapram_segment_end" / Tell,

        "PC_segment" / Tell,
        "PC_offset" / Tell,
        "PC" / Int16ul,
        "PC_offset_end" / Tell,
        "PC_segment_end" / Tell,

        "GTIA_segment" / Tell,
        "GTIA_offset" / Tell,
        "GTIA" / Struct(
            "HPOSP0_offset" / Tell,
            "HPOSP0" / Int8ul,
            "HPOSP1_offset" / Tell,
            "HPOSP1" / Int8ul,
            "HPOSP2_offset" / Tell,
            "HPOSP2" / Int8ul,
            "HPOSP3_offset" / Tell,
            "HPOSP3" / Int8ul,
            "HPOSM0_offset" / Tell,
            "HPOSM0" / Int8ul,
            "HPOSM1_offset" / Tell,
            "HPOSM1" / Int8ul,
            "HPOSM2_offset" / Tell,
            "HPOSM2" / Int8ul,
            "HPOSM3_offset" / Tell,
            "HPOSM3" / Int8ul,
            "PF0PM_offset" / Tell,
            "PF0PM" / Int8ul,
            "PF1PM_offset" / Tell,
            "PF1PM" / Int8ul,
            "PF2PM_offset" / Tell,
            "PF2PM" / Int8ul,
            "PF3PM_offset" / Tell,
            "PF3PM" / Int8ul,
            "M0PL_offset" / Tell,
            "M0PL" / Int8ul,
            "M1PL_offset" / Tell,
            "M1PL" / Int8ul,
            "M2PL_offset" / Tell,
            "M2PL" / Int8ul,
            "M3PL_offset" / Tell,
            "M3PL" / Int8ul,
            "P0PL_offset" / Tell,
            "P0PL" / Int8ul,
            "P1PL_offset" / Tell,
            "P1PL" / Int8ul,
            "P2PL_offset" / Tell,
            "P2PL" / Int8ul,
            "P3PL_offset" / Tell,
            "P3PL" / Int8ul,
            "SIZEP0_offset" / Tell,
            "SIZEP0" / Int8ul,
            "SIZEP1_offset" / Tell,
            "SIZEP1" / Int8ul,
            "SIZEP2_offset" / Tell,
            "SIZEP2" / Int8ul,
            "SIZEP3_offset" / Tell,
            "SIZEP3" / Int8ul,
            "SIZEM_offset" / Tell,
            "SIZEM" / Int8ul,
            "GRAFP0_offset" / Tell,
            "GRAFP0" / Int8ul,
            "GRAFP1_offset" / Tell,
            "GRAFP1" / Int8ul,
            "GRAFP2_offset" / Tell,
            "GRAFP2" / Int8ul,
            "GRAFP3_offset" / Tell,
            "GRAFP3" / Int8ul,
            "GRAFM_offset" / Tell,
            "GRAFM" / Int8ul,
            "COLPM0_offset" / Tell,
            "COLPM0" / Int8ul,
            "COLPM1_offset" / Tell,
            "COLPM1" / Int8ul,
            "COLPM2_offset" / Tell,
            "COLPM2" / Int8ul,
            "COLPM3_offset" / Tell,
            "COLPM3" / Int8ul,
            "COLPF0_offset" / Tell,
            "COLPF0" / Int8ul,
            "COLPF1_offset" / Tell,
            "COLPF1" / Int8ul,
            "COLPF2_offset" / Tell,
            "COLPF2" / Int8ul,
            "COLPF3_offset" / Tell,
            "COLPF3" / Int8ul,
            "COLBK_offset" / Tell,
            "COLBK" / Int8ul,
            "PRIOR_offset" / Tell,
            "PRIOR" / Int8ul,
            "VDELAY_offset" / Tell,
            "VDELAY" / Int8ul,
            "GRACTL_offset" / Tell,
            "GRACTL" / Int8ul,

            "_consol_mask_offset" / Tell,
            "_consol_mask" / Int8ul,
            "_speaker_offset" / Tell,
            "_speaker" / Int32sl,
            "_next_console_value_offset" / Tell,
            "_next_console_value" / Int32sl,
            "_TRIG_latch_offset" / Tell,
            "_TRIG_latch" / Bytes(4),
            "_GTIA_offset_end" / Tell,
            ),
        "GTIA_segment_end" / Tell,

        "PIA_segment" / Tell,
        "PIA_offset" / Tell,
        "PIA" / Struct(
            "PACTL_offset" / Tell,
            "PACTL" / Int8ul,
            "PBCTL_offset" / Tell,
            "PBCTL" / Int8ul,
            "PORTA_offset" / Tell,
            "PORTA" / Int8ul,
            "PORTB_offset" / Tell,
            "PORTB" / Int8ul,

            "_PORTA_mask_offset" / Tell,
            "_PORTA_mask" / Int8ul,
            "_PORTB_mask_offset" / Tell,
            "_PORTB_mask" / Int8ul,

            "_CA2_offset" / Tell,
            "_CA2" / Int32sl,
            "_CA2_negpending_offset" / Tell,
            "_CA2_negpending" / Int32sl,
            "_CA2_pospending_offset" / Tell,
            "_CA2_pospending" / Int32sl,
            "_CB2_offset" / Tell,
            "_CB2" / Int32sl,
            "_CB2_negpending_offset" / Tell,
            "_CB2_negpending" / Int32sl,
            "_CB2_pospending_offset" / Tell,
            "_CB2_pospending" / Int32sl,
            "_PIA_offset_end" / Tell,
            ),
        "PIA_segment_end" / Tell,

        "POKEY_segment" / Tell,
        "POKEY_offset" / Tell,
        "POKEY" / Struct(
            "KBCODE_offset" / Tell,
            "KBCODE" / Int8ul,
            "IRQST_offset" / Tell,
            "IRQST" / Int8ul,
            "IRQEN_offset" / Tell,
            "IRQEN" / Int8ul,
            "SKCTL_offset" / Tell,
            "SKCTL" / Int8ul,

            "_shift_key_offset" / Tell,
            "_shift_key" / Int32sl,
            "_keypressed_offset" / Tell,
            "_keypressed" / Int32sl,
            "_DELAYED_SERIN_IRQ_offset" / Tell,
            "_DELAYED_SERIN_IRQ" / Int32sl,
            "_DELAYED_SEROUT_IRQ_offset" / Tell,
            "_DELAYED_SEROUT_IRQ" / Int32sl,
            "_DELAYED_XMTDONE_IRQ_offset" / Tell,
            "_DELAYED_XMTDONE_IRQ" / Int32sl,

            "AUDF1_offset" / Tell,
            "AUDF1" / Int8ul,
            "AUDF2_offset" / Tell,
            "AUDF2" / Int8ul,
            "AUDF3_offset" / Tell,
            "AUDF3" / Int8ul,
            "AUDF4_offset" / Tell,
            "AUDF4" / Int8ul,
            "AUDC1_offset" / Tell,
            "AUDC1" / Int8ul,
            "AUDC2_offset" / Tell,
            "AUDC2" / Int8ul,
            "AUDC3_offset" / Tell,
            "AUDC3" / Int8ul,
            "AUDC4_offset" / Tell,
            "AUDC4" / Int8ul,
            "AUDCTL_offset" / Tell,
            "AUDCTL" / Int8ul,

            "_DivNIRQ_offset" / Tell,
            "_DivNIRQ" / Bytes(16),
            "_DivNMax_offset" / Tell,
            "_DivNMax" / Bytes(16),
            "_Base_mult_offset" / Tell,
            "_Base_mult" / Int32sl,

            "POKEY_offset_end" / Tell,
            ),
        "POKEY_segment_end" / Tell,

        "Extensions_segment" / Tell,
        "XEP80_enabled_offset" / Tell,
        "XEP80_enabled" / Int8ul,
        "XEP80_offset" / Tell,
        "XEP80" / If(this.XEP80_enabled,
            Struct(
                "tbd_offset" / Tell,
                "tbd" / Bytes(8330),
                # 25 INT,1 => 100 bytes
                # 1 WORD,10 => 10 bytes
                # 3 BYTE, 1 => 3 bytes
                # 1 BYTE, 25 => 25 bytes
                # 8192 video ram
                # ---------
                # 8330 bytes
                #
                # INT(&XEP80_port, 1);
                # INT(&show_XEP80, 1);
                # INT(&num_ticks, 1);
                # INT(&output_word, 1);
                # INT(&input_count, 1);
                # INT(&receiving, 1);
                # UWORD(input_queue, 10);
                # INT(&receiving, 1);
                # UBYTE(&last_char, 1);
                # INT(&xpos, 1);
                # INT(&xscroll, 1);
                # INT(&ypos, 1);
                # INT(&cursor_x, 1);
                # INT(&cursor_y, 1);
                # INT(&curs, 1);
                # INT(&old_xpos, 1);
                # INT(&old_ypos, 1);
                # INT(&lmargin, 1);
                # INT(&rmargin, 1);
                # UBYTE(&attrib_a, 1);
                # UBYTE(&attrib_b, 1);
                # INT(&list_mode, 1);
                # INT(&escape_mode, 1);
                # INT(&char_set, 1);
                # INT(&cursor_on, 1);
                # INT(&cursor_blink, 1);
                # INT(&cursor_overwrite, 1);
                # INT(&blink_reverse, 1);
                # INT(&inverse_mode, 1);
                # INT(&screen_output, 1);
                # INT(&burst_mode, 1);
                # INT(&graphics_mode, 1);
                # INT(&pal_mode, 1);
                # UBYTE(&ptr, 25);
                # UBYTE(video_ram, 8192);
            "XEP80_offset_end" / Tell,
                )
            ),

        "PBI_offset" / Tell,
        "PBI" / Struct(
            "D1FF_LATCH_offset" / Tell,
            "D1FF_LATCH" / Int8ul,
            "D6D7ram_offset" / Tell,
            "D6D7ram" / Int32sl,
            "IRQ_offset" / Tell,
            "IRQ" / Int32sl,
            "PBI_offset_end" / Tell,
            ),

        "PBI_MIO_enabled_offset" / Tell,
        "PBI_MIO_enabled" / Int8ul,
        "PBI_MIO_offset" / Tell,
        "PBI_MIO" / If(this.PBI_MIO_enabled,
            Struct(
                "scsi_disk_filename_offset" / Tell,
                "scsi_disk_filename" / Fname,
                "rom_filename_offset" / Tell,
                "rom_filename" / Fname,
                "ram_size_offset" / Tell,
                "ram_size" / Int32sl,
                "ram_bank_offset_offset" / Tell,
                "ram_bank_offset" / Int32sl,
                "ram_offset" / Tell,
                "ram" / Bytes(this.ram_size),
                "rom_bank_offset" / Tell,
                "rom_bank" / Int8ul,
                "ram_enabled_offset" / Tell,
                "ram_enabled" / Int32sl,
            "PBI_MIO_offset_end" / Tell,
                )
            ),

        "PBI_BB_enabled_offset" / Tell,
        "PBI_BB_enabled" / Int8ul,
        "PBI_BB_offset" / Tell,
        "PBI_BB" / If(this.PBI_BB_enabled,
            Struct(
                "scsi_disk_filename_offset" / Tell,
                "scsi_disk_filename" / Fname,
                "rom_filename_offset" / Tell,
                "rom_filename" / Fname,
                "ram_bank_offset_offset" / Tell,
                "ram_bank_offset" / Int32sl,
                "ram_offset" / Tell,
                "ram" / Bytes(BB_RAM_SIZE),
                "rom_bank_offset" / Tell,
                "rom_bank" / Int8ul,
                "rom_high_bit_offset" / Tell,
                "rom_high_bit" / Int32sl,
                "pcr_offset" / Tell,
                "pcr" / Int8ul,
            "PBI_BB_offset_end" / Tell,
                )
            ),

        "PBI_XLD_enabled_offset" / Tell,
        "PBI_XLD_enabled" / Int8ul,
        "PBI_XLD_offset" / Tell,
        "PBI_XLD" / If(this.PBI_XLD_enabled,
            Struct(
                "xld_v_enabled_offset" / Tell,
                "xld_v_enabled" / Int32sl,
                "xld_d_enabled_offset" / Tell,
                "xld_d_enabled" / Int32sl,
                "xld_d_romfilename_offset" / Tell,
                "xld_d_romfilename" / Fname,
                "xld_v_romfilename_offset" / Tell,
                "xld_v_romfilename" / Fname,
                "votrax_latch_offset" / Tell,
                "votrax_latch" / Int8ul,
                "modem_latch_offset" / Tell,
                "modem_latch" / Int8ul,
                "CommandFrame_offset" / Tell,
                "CommandFrame" / Bytes(COMMAND_FRAME_SIZE),
                "CommandIndex_offset" / Tell,
                "CommandIndex" / Int32sl,
                "DataBuffer_offset" / Tell,
                "DataBuffer" / Bytes(DATA_BUFFER_SIZE),
                "DataIndex_offset" / Tell,
                "DataIndex" / Int32sl,
                "TransferStatus_offset" / Tell,
                "TransferStatus" / Int32sl,
                "ExpectedBytes_offset" / Tell,
                "ExpectedBytes" / Int32sl,
                "VOTRAXSND_busy_offset" / Tell,
                "VOTRAXSND_busy" / Int32sl,
            "PBI_XLD_offset_end" / Tell,
                )
            ),
        "Extensions_segment_end" / Tell,

        )

    return a8save

def get_offsets(container, prefix, comments, segments, dtypes, data_offset=0, current_dtype=None):
    start = 0
    name = "data"
    origin = 0
    for k,v in list(container.items()):
        if k.endswith("_offset"):
            log.debug(("offset", prefix + k[:-7], v))
            comments[v + data_offset] = prefix + k[:-7]
            if current_dtype is not None:
                if len(current_dtype) > 0:
                    dname, dstart, _ = current_dtype.pop()
                    current_dtype.append((dname, dstart, v + data_offset))
                current_dtype.append((k[:-7], v + data_offset, -1))
        elif k.endswith("_offset_end"):
            c = current_dtype
            if c is not None and len(c) > 0:
                dname, dstart, end = c.pop()
                log.debug(("HTOESURCHOSECRUHOEUH", k, dname, dstart, end))
                c.append((dname, dstart, v))
        elif k.endswith("_segment"):
            args = k.split("_")
            args.pop()
            if args[-1].startswith("$"):
                origin = int(args.pop()[1:], 16)
            else:
                origin = 0
            log.debug(("segment:" + prefix + k[:-7], v))
            start = v
            name = " ".join(args)
            current_dtype = []
        elif k.endswith("_segment_end"):
            segments.append((start + data_offset, v + data_offset, origin, name))
            log.debug(("end segment:" + prefix + name, v, segments[-1]))
            if name in dtypes:
                last_dtype = dtypes[name]
                if len(last_dtype) > 0:
                    dname, dstart, _ = last_dtype.pop()
                    last_dtype.append((dname, dstart, v + data_offset))

        elif type(v) == Container:
            log.debug(("Container: %s" % k))
            c = []
            index = get_offsets(v, prefix + "%s_" % k, comments, segments, dtypes, data_offset, c)
            dtypes[k] = c
            log.debug(("End Container: %s" % k, c))


def parse_state(data, data_offset):
    a8save = init_atari800_struct()
    test = a8save.parse(data)
    offsets = {}
    segments = []
    dtypes_setup = {}
    get_offsets(test, "", offsets, segments, dtypes_setup, data_offset)
    dtypes = {}
    for name, index_data in dtypes_setup.items():
        d = []
        log.debug(("dtypes for %s" % name))
        for varname, start, stop in index_data:
            log.debug(("varname", varname, start, stop))
            size = stop - start
            if size == 1:
                d.append((varname, np.uint8))
            elif size == 2:
                d.append((varname, '>u2'))
            elif size == 4:
                d.append((varname, '>i4'))
            else:
                size = stop - start
                if size < 0:
                    size = 1
                d.append((varname, np.uint8, size))
        log.debug(("dtype=%s" % str(d)))
        dtypes[name] = np.dtype(d)
    names = {v: k for k, v in offsets.items()}
    segment_starts = {name: start for start, _, _, name in segments}
    return offsets, names, segments, segment_starts, dtypes


if __name__ == "__main__":
    import sys
    with open(sys.argv[1], "rb") as fh:
        data = fh.read()
        offsets, names, segments, segment_starts, dtypes = parse_state(data, 0)
        # print sorted(list(offsets.items()))
        # print "\n".join(["%s: %s" % kv for kv in sorted(list(names.items()))])
        print(dtypes['ANTIC'])
        print(segment_starts['ANTIC'])


"""Debugger interface mimicking GDB

info|i
  breakpoints (bp)   List breakpoints
  cpu  show CPU registers
  antic
  pokey
  gtia


break|b ADDRESS|LABEL|+OFFSET|-OFFSET [if CONDITION]
  ADDRESS:
    literal hex address at which to break
  LABEL:
    break at address stored in LABEL
  +OFFSET
    break at current address + offset
  -OFFSET
    break at current address - offset
  CONDITION:
    an expression that evaluates to a boolean, triggering the breakpoint when
    true.  E.g. (A > 10 and (X < 0x80 and not C))


watch|w CONDITION
  CONDITION:
    same as above (watchpoints are the same as breakpoints without an address)


tbreak|t COMMAND [if CONDITION]
  Temporary break: break once only and then it is removed. See "break" above.


delete|d [BREAKPOINT[-RANGE]]
  Delete breakpoints

  BREAKPOINT:
    delete specified breakpoint
  RANGE:
    delete range from BREAKPOINT to RANGE

  No argument: delete all breakpoints


disable|di [BREAKPOINT[-RANGE]]
  Disable breakpoints, specifiers as above


enable|en [BREAKPOINT[-RANGE]]
  Enable breakpoints, specifiers as above


continue|c [COUNT]
  Continue executing until next breakpoint

  COUNT:
    Continue, but ignore current breakpoint NUMBER times


finish|f
  Continue to end of function


step|s [NUMBER]
  Step to next line, will step into functions

  NUMBER:
    number of steps to perform


next|n [NUMBER]
  Next line, stepping over function calls


until|u [ADDRESS|LABEL]
  Continue until reaching address


NON-GDB commands:

ccontinue|cc NUMBER
  Continue executing under NUMBER cycles has been reached.


"""

import numpy as np

from . import dtypes as dd
from ..utils.persistence import Serializable

import logging
log = logging.getLogger(__name__)


class Breakpoint:
    def __init__(self, debugger, id, addr=None):
        self.debugger = debugger
        self.id = id
        self.index = id * dd.TOKENS_PER_BREAKPOINT
        if addr is not None:
            self.simple_address(addr)

    def __str__(self):
        if self.id == 0:
            ref = self.reference_value
        else:
            ref = "n/a"
        return f"<breakpoint {self.id}, status={self.status}, type={self.type}, ref_val={ref} terms={self.terms}>"

    @property
    def status(self):
        c = self.debugger.debug_cmd[0]
        return c['breakpoint_status'][self.id]

    @status.setter
    def status(self, status):
        c = self.debugger.debug_cmd[0]
        c['breakpoint_status'][self.id] = status

    @property
    def type(self):
        c = self.debugger.debug_cmd[0]
        return c['breakpoint_type'][self.id]

    @type.setter
    def type(self, type):
        c = self.debugger.debug_cmd[0]
        c['breakpoint_type'][self.id] = type

    @property
    def reference_value(self):
        c = self.debugger.debug_cmd[0]
        return c['reference_value'][self.id]

    @reference_value.setter
    def reference_value(self, type):
        c = self.debugger.debug_cmd[0]
        c['reference_value'][self.id] = reference_value

    @property
    def terms(self):
        c = self.debugger.debug_cmd[0]
        i = self.index
        tokens = c['tokens'][i:i+dd.TOKENS_PER_BREAKPOINT]
        end_tokens = np.where(tokens == dd.END_OF_LIST)[0]
        return tokens[:end_tokens[0]]

    @terms.setter
    def terms(self, term_list):
        c = self.debugger.debug_cmd[0]
        i = self.index
        count = len(term_list)
        c['tokens'][i:i+count] = term_list

    @property
    def enabled(self):
        c = self.debugger.debug_cmd[0]
        return bool(c['breakpoint_status'][self.id] == dd.BREAKPOINT_ENABLED)

    @property
    def had_error(self):
        c = self.debugger.debug_cmd[0]
        return bool(c['breakpoint_status'][self.id] & dd.BREAKPOINT_ERROR)

    def simple_address(self, addr):
        # shortcut to create a PC=addr breakpoint
        c = self.debugger.debug_cmd[0]
        c['breakpoint_type'][self.id] = dd.BREAKPOINT_CONDITIONAL
        c['breakpoint_status'][self.id] = dd.BREAKPOINT_ENABLED
        self.terms = (dd.REG_PC, dd.NUMBER, addr, dd.OP_EQ, dd.END_OF_LIST)

    def step_into(self, count):
        # shortcut to create a break after `count` instructions
        c = self.debugger.debug_cmd[0]
        c['breakpoint_type'][self.id] = dd.BREAKPOINT_COUNT_INSTRUCTIONS
        c['breakpoint_status'][self.id] = dd.BREAKPOINT_ENABLED
        c['reference_value'][self.id] = self.debugger.instructions_since_power_on
        c['tokens'][self.index] = count
        self.enable()

    def break_scan_line(self, count):
        # shortcut to create a break at the next (or a subsequent) scan line
        c = self.debugger.debug_cmd[0]
        c['breakpoint_type'][self.id] = dd.BREAKPOINT_COUNT_LINES
        c['breakpoint_status'][self.id] = dd.BREAKPOINT_ENABLED
        c['reference_value'][self.id] = self.debugger.scan_lines_since_power_on
        c['tokens'][self.index] = count
        self.enable()

    def break_vbi_start(self, count):
        # shortcut to create a break at the next vertical blank interrupt
        c = self.debugger.debug_cmd[0]
        c['breakpoint_type'][self.id] = dd.BREAKPOINT_CONDITIONAL
        c['breakpoint_status'][self.id] = dd.BREAKPOINT_ENABLED
        c['tokens'][self.index] = count
        self.terms = (dd.EMU_VBI_START, dd.END_OF_LIST)
        self.enable()

    def break_vbi_end(self, count):
        # shortcut to create a break at the next vertical blank interrupt
        c = self.debugger.debug_cmd[0]
        c['breakpoint_type'][self.id] = dd.BREAKPOINT_CONDITIONAL
        c['breakpoint_status'][self.id] = dd.BREAKPOINT_ENABLED
        c['tokens'][self.index] = count
        self.terms = (dd.EMU_VBI_END, dd.END_OF_LIST)
        self.enable()

    def break_dli_start(self, count):
        # shortcut to create a break at the next display list interrupt
        c = self.debugger.debug_cmd[0]
        c['breakpoint_type'][self.id] = dd.BREAKPOINT_CONDITIONAL
        c['breakpoint_status'][self.id] = dd.BREAKPOINT_ENABLED
        c['tokens'][self.index] = count
        self.terms = (dd.EMU_DLI_START, dd.END_OF_LIST)
        self.enable()

    def break_dli_end(self, count):
        # shortcut to create a break at the next display list interrupt
        c = self.debugger.debug_cmd[0]
        c['breakpoint_type'][self.id] = dd.BREAKPOINT_CONDITIONAL
        c['breakpoint_status'][self.id] = dd.BREAKPOINT_ENABLED
        c['tokens'][self.index] = count
        self.terms = (dd.EMU_DLI_END, dd.END_OF_LIST)
        self.enable()

    def count_cycles(self, count):
        # shortcut to create a break after `count` machine cycles
        c = self.debugger.debug_cmd[0]
        c['breakpoint_type'][self.id] = dd.BREAKPOINT_COUNT_CYCLES
        c['breakpoint_status'][self.id] = dd.BREAKPOINT_ENABLED
        c['reference_value'][self.id] = self.debugger.cycles_since_power_on
        c['tokens'][self.index] = count
        self.enable()

    def count_frames(self, count):
        # shortcut to create a break after `count` instructions
        c = self.debugger.debug_cmd[0]
        c['breakpoint_type'][self.id] = dd.BREAKPOINT_COUNT_FRAMES
        c['breakpoint_status'][self.id] = dd.BREAKPOINT_ENABLED
        c['reference_value'][self.id] = self.debugger.current_frame_number
        c['tokens'][self.index] = count
        self.enable()

    def break_at_return(self):
        # shortcut to create a PC=addr breakpoint
        c = self.debugger.debug_cmd[0]
        c['breakpoint_type'][self.id] = dd.BREAKPOINT_CONDITIONAL
        c['breakpoint_status'][self.id] = dd.BREAKPOINT_ENABLED
        sp = self.debugger.stack_pointer
        self.terms = (dd.OPCODE_TYPE, dd.OPCODE_RETURN, dd.REG_SP, dd.NUMBER, sp, dd.OP_EQ, dd.OP_LOGICAL_AND, dd.END_OF_LIST)
        self.enable()

    def clear(self):
        c = self.debugger.debug_cmd[0]
        status = dd.BREAKPOINT_DISABLED if self.id == 0 else dd.BREAKPOINT_EMPTY
        c['breakpoint_type'][self.id] = dd.BREAKPOINT_CONDITIONAL
        c['breakpoint_status'][self.id] = status
        c['tokens'][self.index] = dd.END_OF_LIST

    def enable(self):
        c = self.debugger.debug_cmd[0]
        c['breakpoint_status'][self.id] = dd.BREAKPOINT_ENABLED
        if self.id >= c['num_breakpoints']:
            c['num_breakpoints'] = self.id + 1

    def disable(self):
        c = self.debugger.debug_cmd[0]
        c['breakpoint_status'][self.id] = dd.BREAKPOINT_DISABLED


class Debugger(Serializable):
    name = "<name>"

    serializable_attributes = ['debug_cmd_raw']
    serializable_computed = {'debug_cmd_raw'}

    def __init__(self):
        self.debug_cmd_raw = np.zeros([dd.DEBUGGER_COMMANDS_DTYPE.itemsize], dtype=np.uint8)
        self.debug_cmd = self.debug_cmd_raw.view(dtype=dd.DEBUGGER_COMMANDS_DTYPE)
        self.clear_all_breakpoints()

    ##### Serialization

    def restore_computed_attributes(self, state):
        self.debug_cmd_raw[:] = state['debug_cmd_raw']

    def clear_all_breakpoints(self):
        c = self.debug_cmd[0]
        c['breakpoint_type'][:] = dd.BREAKPOINT_CONDITIONAL
        c['breakpoint_status'][:] = 0
        c['breakpoint_status'][0] = dd.BREAKPOINT_DISABLED
        c['num_breakpoints'] = 0
        c['last_pc'] = -1

    def create_breakpoint(self, addr=None):
        c = self.debug_cmd[0]
        empty = np.where(c['breakpoint_status'] == dd.BREAKPOINT_EMPTY)[0]
        bpid = empty[0]
        c['num_breakpoints'] = max(c['num_breakpoints'], bpid + 1)
        return Breakpoint(self, bpid, addr)

    def get_breakpoint(self, bpid):
        if bpid < 0:
            return None
        return Breakpoint(self, bpid)

    def step_into(self, number=1):
        b = Breakpoint(self, 0)
        b.step_into(number)

    count_instructions = step_into

    def break_scan_line(self, number=1):
        b = Breakpoint(self, 0)
        b.break_scan_line(number)

    def break_vbi_start(self, number=1):
        b = Breakpoint(self, 0)
        b.break_vbi_start(number)

    def break_vbi_end(self, number=1):
        b = Breakpoint(self, 0)
        b.break_vbi_end(number)

    def break_dli_start(self, number=1):
        b = Breakpoint(self, 0)
        b.break_dli_start(number)

    def break_dli_end(self, number=1):
        b = Breakpoint(self, 0)
        b.break_dli_end(number)

    def count_frames(self, number=1):
        b = Breakpoint(self, 0)
        b.count_frames(number)

    def count_cycles(self, cycles=1):
        b = Breakpoint(self, 0)
        b.count_cycles(number)

    def iter_breakpoints(self):
        for i in range(dd.NUM_BREAKPOINT_ENTRIES):
            b = Breakpoint(self, i)
            if b.enabled:
                yield b

    def print_breakpoints(self):
        for b in self.iter_breakpoints():
            print(b)



import os
import inspect
import pkg_resources

from . import errors

import logging
log = logging.getLogger(__name__)


class Compressor:
    """Compressor (packer/unpacker) for disk image compression.

    In their native data format, disk images may be stored as raw data or can
    be compressed by any number of techniques. Subclasses of Compressor
    implement the `calc_unpack_data` method which examines the byte_data
    argument for the supported compression type, and if valid returns the
    unpacked bytes to be used in the disk image parsing.

    Compressors are stateless, but are implemented as normal classes for
    convenience and ease of subclassing.
    """
    compression_algorithm = None

    def __init__(self, byte_data=None):
        if byte_data is not None:
            self.unpacked = self.calc_unpacked_data(byte_data)

    def __str__(self):
        return self.compression_algorithm

    @property
    def is_compressed(self):
        return self.compression_algorithm != Uncompressed.compression_algorithm

    #### decompression

    def calc_unpacked_data(self, byte_data):
        """Attempt to unpack `byte_data` using this unpacking algorithm.

        `byte_data` must be a byte array or something that can pose for a byte
        array.

        If the data is not recognized by this subclass, raise an
        `InvalidAlgorithm` exception. The caller must recognize this and move
        on to trying the next compressor in the list.

        If the data is recognized by this subclass but the unpacking algorithm
        is not implemented, raise an `UnsupportedAlgorithm` exception. This
        is different than the `InvalidAlgorithm` exception because it
        indicates that the data was indeed recognized by this subclass (despite
        not being unpacked) and checking further compressors is not necessary.
        """
        raise errors.InvalidAlgorithm(f"Uncompressing '{self.compression_algorithm}' not implemented")

    #### compression

    def calc_packed_data(self, byte_data, media=None):
        """Pack this data into a compressed data array using this packing
        algorithm.

        Subclasses should raise the `UnsupportedAlgorithm` exception if
        compression is not implemented.

        `media` is supplied in case the compression format needs the disk
        geometry, e.g. in DCM compression.
        """
        raise errors.InvalidAlgorithm(f"Compression for '{self.compression_algorithm}' not implemented.")


class Uncompressed(Compressor):
    compression_algorithm = "none"

    def calc_unpacked_data(self, byte_data):
        return byte_data

    def calc_packed_data(self, byte_data, media):
        return byte_data


_compressors = None

def _find_compressors():
    compressors = []
    for entry_point in pkg_resources.iter_entry_points('atrip.compressors'):
        mod = entry_point.load()
        log.debug(f"find_compressor: Found module {entry_point.name}={mod.__name__}")
        for name, obj in inspect.getmembers(mod):
            if inspect.isclass(obj) and Compressor in obj.__mro__[1:]:
                log.debug(f"find_compressors:   found compressor class {name}")
                compressors.append(obj)
    return compressors

def find_compressors():
    global _compressors

    if _compressors is None:
        _compressors = _find_compressors()
    return _compressors

def find_compressor_by_name(name):
    items = find_compressors()
    for c in items:
        if c.compression_algorithm == name:
            return c()
    raise KeyError(f"Unknown compressor {name}")

def guess_compressor(raw_data):
    compressor = None
    for c in find_compressors():
        log.debug(f"trying compressor {c.compression_algorithm}")
        try:
            compressor = c(raw_data)
        except (errors.InvalidAlgorithm, IOError, EOFError) as e:
            continue
        else:
            log.info(f"found compressor {c.compression_algorithm}")
            break
    else:
        log.debug(f"image does not appear to be compressed.")
        compressor = Uncompressed(raw_data)
    return compressor

def guess_compressor_list(data):
    compressors = []
    while True:  # loop until reach an uncompressed state
        c = guess_compressor(data)
        data = c.unpacked
        if not c.is_compressed:
            if not compressors:
                # save the null compressor only if it's the only one
                log.info(f"image does not appear to be compressed.")
                compressors.append(c.__class__)
            break
        compressors.append(c.__class__)
    return data, compressors

def compress_in_reverse_order(byte_data, decompression_order, media=None, skip_missing_compressors=False):
    order = reversed(decompression_order)
    log.debug(f"compression_order: {order}")
    for compressor_cls in order:
        compressor = compressor_cls()
        try:
            byte_data = compressor.calc_packed_data(byte_data, media)
        except errors.InvalidAlgorithm:
            if skip_missing_compressors:
                continue
            else:
                log.error(f"Compression algorithm {compressor} not yet implemented")
                raise
    return byte_data



import numpy as np

from .. import errors
from ..stringifier import Stringifier


class Hexify(Stringifier):
    text_type = "hexify"
    ui_name = "Hex String"

    def calc_text(self, byte_data):
        text = " ".join(["%02x" % i for i in byte_data])
        return text



# Standard library imports.
import sys
import os

# Major package imports.
import wx
import numpy as np
import json

from sawx.editor import SawxEditor
from sawx.ui.tilemanager import TileManager
from sawx.persistence import get_template
from sawx.events import EventHandler

from ..viewer import find_viewer_class_by_name

import logging
log = logging.getLogger(__name__)


class DummyLinkedBase(object):
    segment = None
    segment_uuid = None

class DummyFocusedViewer(object):
    linked_base = DummyLinkedBase


class TileManagerBase(SawxEditor):
    """Base class for editors that use the tile manager as the central window
    manager

    The TileManager splits up the central window into a set of viewers that ar
    arranged in a layout in relation to one another. This layout can be saved
    and restored along with the data being edited.

    The layout can also be overridden when loading an editor. The layout/viewer
    creation priority is as follows:

    * viewers specified on the command line (any saved layout will be ignored)
    * viewers & layout restored from the save file
    * viewers & layout from a template file based on a combination of the
      editor and MIME type of the file
    * viewers specified in the default_viewers class attribute of the editor
    """
    editor_id = "omnivore.tile_manager_base"
    ui_name = "Tile Manager Base"

    default_viewers = "dummy"

    def prepare_destroy(self):
        self.document.halt_background_processing()
        self.focused_viewer = None
        # Operate on copy of list because you can't iterate and remove from the
        # same list
        for v in list(self.viewers):
            log.debug(f"Closing viewer: {v}")
            self.viewers.remove(v)
            v.prepare_for_destroy()
            del v

    def create_event_bindings(self):
        self.focused_viewer = None
        self.focused_viewer_changed_event = EventHandler(self)
        self.viewers = []

    def create_control(self, parent):
        """Creates the `TileManager` control that will manage all the viewers
        """

        panel = TileManager(parent, toggle_checker=self.check_viewer_center_base)
        panel.Bind(TileManager.EVT_CLIENT_ACTIVATED, self.on_viewer_active)
        panel.Bind(TileManager.EVT_CLIENT_CLOSE, self.on_viewer_close)
        panel.Bind(TileManager.EVT_CLIENT_REPLACE, self.on_viewer_replace)
        panel.Bind(TileManager.EVT_CLIENT_TOGGLE_REQUESTED, self.on_viewer_link)
        return panel

    def has_command_line_viewer_override(self, args):
        return bool(args)

    @property
    def layout_template_search_order(self):
        """Override this to return a list of better guesses for a default
        layout based on the current state of the editor
        """
        return [self.editor_id]

    def get_default_layout(self, template_name=None):
        if template_name is None:
            template_name = self.default_layout_template_name
        template_name = self.document.calc_layout_template_name(template_name)
        log.debug("template from: %s" % template_name)
        try:
            data = get_template(template_name)
        except OSError as err:
            log.error(f"Failed loading template: {err}")
            layout = {}
        else:
            try:
                layout = json.loads(data)
            except ValueError:
                log.error("invalid data in default layout")
                layout = {}
        return layout

    def calc_viewers_from_layout(self, s):
        viewers = s.get("viewers", {})
        if viewers:
            layout = s.get("layout", {})
            if layout:
                self.control.restore_layout(layout)
        return viewers

    def restore_layout(self, s, force=False):
        """Create a layout of viewers, either from the saved layout in the
        session file, or by using a list of template names to find a default
        layout.

        Returns: the layout session dict used
        """
        log.debug(f"restore_layout: trying saved layout")
        viewers = self.calc_viewers_from_layout(s)
        if not viewers:
            for template_name in self.layout_template_search_order:
                log.debug(f"restore_layout: trying template {template_name}")
                s = self.get_default_layout(template_name)
                viewers = self.calc_viewers_from_layout(s)
                if viewers:
                    log.debug(f"restore_layout: using template {template_name}")
                    break
        if not viewers:
            log.debug(f"restore_layout: no template {default_name}, falling back to {self.default_viewers}")
            viewers = [{'name':name, 'uuid':name} for name in self.default_viewers.split(",")]
        log.debug(f"restore_layout: viewers={viewers}")

        if "initial_segment_type" in s:
            self.change_initial_segment(s["initial_segment_type"])

        viewer_metadata = {}
        for v in viewers:
            viewer_metadata[v['uuid']] = v
            log.debug("metadata: viewer[%s]=%s" % (v['uuid'], str(v)))

        self.create_viewers(viewer_metadata, force)

        return s

    def replace_layout(self, s):
        """Replace the existing layout with an entirely new layout.
        """
        log.debug(f"replace_layout: replacing current layout")
        self.control.remove_all()
        self.restore_layout(s, True)

    def change_initial_segment(self, segment_ui_name):
        """Change the initial viewed segment to the first one that matches
        the ui_name
        """
        pass

    def set_initial_focused_viewer(self):
        if self.focused_viewer is None:
            for viewer in self.viewers:
                if not self.control.in_sidebar(viewer.control):
                    break
            print(("setting focus to %s" % viewer))
            self.set_focused_viewer(viewer)
            self.force_focus(viewer)

    def serialize_session(self, s, save_linked_bases=True):
        s["layout"] = self.control.calc_layout()
        s["viewers"] = []
        bases = {}
        for v in self.viewers:
            b = v.linked_base
            bases[b.uuid] = b
            if save_linked_bases:
                e = {"linked base": v.linked_base.uuid}
            else:
                e = {}
            v.serialize_session(e)
            s["viewers"].append(e)
        s["focused viewer"] = self.focused_viewer.uuid
        if save_linked_bases:
            if self.center_base is not None:
                bases[self.center_base.uuid] = self.center_base
                s["center_base"] = self.center_base.uuid
            else:
                s["center_base"] = None
            s["linked bases"] = []
            for u, b in bases.items():
                e = {}
                b.serialize_session(e)
                s["linked bases"].append(e)

    #### viewer utilities

    def reconfigure_viewers(self):
        self.update_viewer_names()

    def update_viewer_names(self):
        for viewer in self.viewers:
            viewer.update_caption()
        self.control.update_captions()

    def add_viewer(self, viewer_or_viewer_cls, linked_base=None, replace_uuid=None):
        if hasattr(viewer_or_viewer_cls, "control"):
            viewer = viewer_or_viewer_cls
        else:
            if linked_base is None:
                if self.focused_viewer is not None:
                    linked_base = self.focused_viewer.linked_base
                else:
                    raise RuntimeError("Creating a viewer with no linked base and no focused viewer")
            viewer = viewer_or_viewer_cls.viewer_factory(self.control, linked_base)
        print("VIWER", viewer, type(viewer_or_viewer_cls), linked_base)
        print("SEGMENT", linked_base.segment)
        if replace_uuid:
            viewer.uuid = replace_uuid
        self.viewers.append(viewer)
        self.control.add(viewer.control, viewer.uuid)
        viewer.recalc_data_model()
        self.update_pane_names()
        return viewer

    def add_viewer_by_name(self, viewer_name):
        viewer_cls = find_viewer_class_by_name(viewer_name)
        self.add_viewer(viewer_cls)

    def replace_viewer(self, viewer_to_replace, new_viewer, linked_base):
        # self.viewers.remove(viewer_to_replace)
        # viewer_to_replace.prepare_for_destroy()
        created_viewer = self.add_viewer(new_viewer, linked_base, replace_uuid=viewer_to_replace.uuid)
        # self.set_focused_viewer(created_viewer)
        # del viewer_to_replace

    def replace_focused_viewer(self, linked_base):
        viewer_cls = self.focused_viewer.__class__
        self.replace_viewer(self.focused_viewer, viewer_cls, linked_base)

    def verify_center_base_is_used(self):
        if self.center_base is not None:
            for v in self.viewers:
                if v.linked_base == self.center_base:
                    break
            else:
                self.center_base = None
        print(f"Center base is: {self.center_base}")

    def unlink_viewer(self):
        number = self.focused_viewer.linked_base.segment_uuid
        base = LinkedBase(editor=self)
        base.view_segment_uuid(number)
        self.replace_focused_viewer(base)
        self.verify_center_base_is_used()

    def link_viewer(self):
        if self.center_base is None:
            self.center_base = self.focused_viewer.linked_base
        else:
            base = self.center_base
            self.replace_focused_viewer(base)

    def check_viewer_center_base(self, viewer_control, toggle_id):
        try:
            v = viewer_control.segment_viewer
        except AttributeError:
            state = True
        else:
            state = v.linked_base == self.center_base
        return state

    def calc_linked_base_from_viewer_metadata(self, e):
        try:
            linked_base = self.linked_bases[e['linked base']]
        except KeyError:
            try:
                name = e['segment_name']
            except KeyError:
                linked_base = self.center_base
            else:
                try:
                    linked_base = self.calc_linked_base_by_name(name)
                except KeyError:
                    linked_base = self.center_base
                else:
                    self.linked_bases[linked_base.uuid] = linked_base
        log.debug(f"calc_linked_base_from_viewer_metadata: {linked_base}")
        return linked_base

    def verify_linked_base(self, linked_base):
        """Make sure linked base is recorded in list"""
        if linked_base.uuid not in self.linked_bases:
            self.linked_base[linked_base.uuid] = linked_base

    def create_viewers(self, viewer_metadata, force=False):
        # Create a set of viewers from a list
        import pprint
        log.debug("viewer_metadata: %s" % str(list(viewer_metadata.keys())))

        #self.document.find_initial_visible_segment(self.center_base)

        layer = 0
        viewers = list(viewer_metadata.keys())

        center_base_used = False
        if not self.viewers or force:
            for uuid in viewers:
                log.debug("loading viewer: %s" % uuid)
                e = viewer_metadata[uuid]
                if e:
                    viewer_type = e['name']
                else:  # either not a uuid or an unknown uuid
                    viewer_type = uuid  # try the value of 'uuid' as a viewer name
                try:
                    viewer_cls = find_viewer_class_by_name(viewer_type)
                except ValueError:
                    log.error("unknown viewer %s, uuid=%s" % (viewer_type, uuid))
                    continue
                log.debug("identified viewer: %s" % viewer_cls)

                if e:
                    linked_base = self.calc_linked_base_from_viewer_metadata(e)
                    log.debug("recreating viewer %s: %s" % (viewer_type, uuid))
                else:  # either not a uuid or an unknown uuid
                    linked_base = viewer_cls.calc_segment_specific_linked_base(self)
                    if linked_base is None:
                        linked_base = self.center_base
                    log.debug("using default metadata for %s" % (viewer_type))
                if linked_base == self.center_base:
                    center_base_used = True

                log.debug("creating viewer %s (%s) with linked base %s" % (uuid, viewer_type, str(linked_base)))
                viewer = viewer_cls.viewer_factory(self.control, linked_base, uuid, e)
                log.debug("created viewer %s (%s)" % (viewer.uuid, viewer.name))

                self.viewers.append(viewer)
                if not self.control.replace_by_uuid(viewer.control, viewer.uuid):
                    log.debug("viewer %s not found, adding in new pane" % viewer.uuid)
                    self.control.add(viewer.control, viewer.uuid)

            if not self.viewers:
                # just load default hex editor if nothing has been created
                viewers = ['hex']
                first = False

        if not center_base_used:
            self.center_base = None
        self.update_pane_names()

    def find_viewer_by_uuid(self, u):
        for v in self.viewers:
            if u == v.uuid:
                return v
        return None

    #### wx event handlers

    def force_focus(self, viewer):
        self.control.force_focus(viewer.uuid)
        c = viewer.control
        if viewer.has_caret and not c.caret_handler.has_carets:
            c.caret_handler.move_current_caret_to_index(c.table, 0)
            flags = c.create_mouse_event_flags()
            flags.source_control = c
            flags.sync_caret_from_control = c
            self.linked_base.sync_caret_event(flags=flags)
            wx.CallAfter(c.SetFocus)
        self.update_pane_names()
        viewer.update_toolbar()

    def set_focused_viewer(self, viewer):
        if self.focused_viewer != viewer:
            if self.focused_viewer is not None:
                self.focused_viewer.lost_focus()
            self.focused_viewer = viewer
            self.focused_viewer_changed_event(viewer)
            self.caret_handler = viewer.linked_base
            viewer.linked_base.segment_selected_event(viewer.linked_base.segment_uuid)
            self.set_menu_for_viewer(viewer)

    def set_menu_for_viewer(self, viewer):
        """Update menubar/toolbar for the current viewer.

        Viewers can have menu items that only apply when that viewer is
        currently focused, so this will change the UI every time a new viewer
        is selected.

        As currently implemented, all possible menu/tool items are listed in
        the class attributes `menubar_desc` and `toolbar_desc`, and the viewer
        prunes the list to remove items that aren't applicable. However, the
        entire description could be built on the fly by rewriting the call to
        `SegmentViewer.prune_menubar` to build a new description list from
        scratch.
        """
        # Override the class attributes with instance attributes
        self.menubar_desc = viewer.prune_menubar(self.__class__.menubar_desc)
        self.toolbar_desc = viewer.prune_toolbar(self.__class__.toolbar_desc)

        # force rebuild of UI
        self.frame.make_active(self, True)

    def on_viewer_active(self, evt):
        try:
            v = evt.child.segment_viewer
        except AttributeError:
            # must be an empty window (a multisash window that has no segment
            # viewer). It can be closed without any further action.
            pass
        else:
            v = evt.child.segment_viewer
            if v == self.focused_viewer:
                log.debug("on_viewer_active: already current viewer %s" % v)
            else:
                log.debug("on_viewer_active: activated viewer %s %s" % (v, v.window_title))
                self.set_focused_viewer(v)

    def on_viewer_close(self, evt):
        log.debug(f"on_viewer_close: child window {evt.child}")
        try:
            v = evt.child.segment_viewer
        except AttributeError:
            # must be an empty window (a multisash window that has no segment
            # viewer). It can be closed without any further action.
            pass
        else:
            log.debug("on_viewer_close: closed viewer %s %s" % (v, v.window_title))

            # Keep a reference to the linked base
            linked_base_save = v.linked_base

            self.viewers.remove(v)
            v.prepare_for_destroy()

            from .. import viewer as omnivore8bit_viewers
            if not self.viewers:
                v = self.add_viewer(omnivore8bit_viewers.PlaceholderViewer, linked_base_save)
            self.set_focused_viewer(self.viewers[0])
            del v

    def on_viewer_replace(self, evt):
        log.debug(f"on_viewer_replace: child window {evt.child}")
        try:
            v = evt.child.segment_viewer
        except AttributeError:
            # must be an empty window (a multisash window that has no segment
            # viewer). It can be closed without any further action.
            pass
        else:
            log.debug("on_viewer_replace: closing viewer %s %s for replacement" % (v, v.window_title))

            # Keep a reference to the linked base
            linked_base_save = v.linked_base

            self.viewers.remove(v)
            v.prepare_for_destroy()
            try:
                new_viewer = evt.replacement_child.segment_viewer
            except AttributeError:
                # New viewer is empty window
                pass
            else:
                self.set_focused_viewer(new_viewer)
            del v

    def on_viewer_link(self, evt):
        try:
            v = evt.child.segment_viewer
        except AttributeError:
            # must be an empty window (a multisash window that has no segment
            # viewer). It can be closed without any further action.
            pass
        else:
            toggle = evt.GetInt()
            desired_state = evt.IsChecked()
            log.debug("on_viewer_replace: linking viewer %s %s: %s" % (v, v.window_title, desired_state))
            if desired_state:
                print("LINKED!")
                self.link_viewer()
            else:
                print("UNLINKED!")
                self.unlink_viewer()



import re

import numpy as np

from sawx.utils.parseutil import NumpyIntExpression, ParseException

import logging
log = logging.getLogger(__name__)


class BaseSearcher(object):
    ui_name = "<base class>"

    def __init__(self, editor, search_text, search_copy, **kwargs):
        self.search_text = self.get_search_text(search_text)
        self.search_copy = search_copy
        if len(self.search_text) > 0:
            self.matches = self.get_matches(editor)
        else:
            self.matches = []

    def get_search_text(self, text):
        return bytes(text, "utf-8")

    def get_matches(self, editor):
        text = self.search_copy
        rs = re.escape(bytes(self.search_text))
        matches = [(i.start(), i.end()) for i in re.finditer(rs, text)]
        return matches


class HexSearcher(BaseSearcher):
    ui_name = "hex"

    def __str__(self):
        return "hex matches: %s" % str(self.matches)

    def get_search_text(self, text):
        try:
            return bytes.fromhex(text)
        except ValueError:
            log.debug("%s: fromhex failed on: %s")
            return ""


class CharSearcher(BaseSearcher):
    ui_name = "text"

    def __str__(self):
        return "char matches: %s" % str(self.matches)


class CommentSearcher(BaseSearcher):
    ui_name = "comments"

    def __str__(self):
        return "comment matches: %s" % str(self.matches)

    def get_search_text(self, text):
        return text

    def get_matches(self, editor):
        segment = editor.segment
        s = segment.origin
        matches = []
        if editor.last_search_settings.get('match_case', False):
            search_text = self.search_text
            for index, comment in segment.iter_comments_in_segment():
                if search_text in comment:
                    matches.append((index, index + 1))
        else:
            search_text = self.search_text.lower()
            for index, comment in segment.iter_comments_in_segment():
                if search_text in comment.lower():
                    matches.append((index, index + 1))
        return matches


class AlgorithmSearcher(BaseSearcher):
    def __str__(self):
        return "pyparsing matches: %s" % str(self.matches)

    def get_search_text(self, text):
        return text

    def get_matches(self, editor):
        s = editor.segment
        a = np.arange(s.origin, s.origin + len(s))
        b = np.asarray(self.search_copy, dtype=np.uint8)
        v = {
            'a': a,
            'b': b,
            }
        expression = NumpyIntExpression(v)
        try:
            result = expression.eval(self.search_text)
            matches = s.bool_to_ranges(result)
            return matches
        except ParseException as e:
            raise ValueError(e)


known_searchers = [
    HexSearcher,
#    CharSearcher,
#    DisassemblySearcher,
#    CommentSearcher,
    ]



"""Byte modification actions
"""
import os
import sys

import wx

from sawx.ui.dialogs import prompt_for_hex, prompt_for_slice

from ..action import ViewerAction
from ..commands import byte_editor as commands

import logging
log = logging.getLogger(__name__)


class IndexRangeAction(ViewerAction):
    """base class for byte modification operations
    """
    doc_hint = "parent,list,level 3"
    parent_doc = """The commands in this menu operate on the current selection to change the byte values to:
    """
    cmd = None

    def calc_name(self, action_key):
        return self.cmd.ui_name

    def calc_enabled(self, action_key):
        return self.editor.can_copy

    def get_cmd(self, editor, segment, ranges):
        return self.cmd(segment, ranges)

    def perform(self, action_key):
        e = self.editor
        cmd = self.get_cmd(e, e.segment, e.current_selection)
        e.process_command(cmd)


class byte_set_to_zero(IndexRangeAction):
    """Set to zero"""
    cmd = commands.ZeroCommand


class byte_set_to_ff(IndexRangeAction):
    """Set to $ff"""
    cmd = commands.FFCommand


class byte_nop(IndexRangeAction):
    """Set to the NOP command of the current processor"""
    cmd = commands.NOPCommand

    def calc_enabled(self, action_key):
        return self.viewer.has_cpu

    def get_cmd(self, editor, segment, ranges):
        nop = self.viewer.machine.get_nop()
        return self.cmd(segment, ranges, nop)


class byte_set_high_bit(IndexRangeAction):
    """Or with $80"""
    cmd = commands.SetHighBitCommand


class byte_clear_high_bit(IndexRangeAction):
    """And with $7f"""
    cmd = commands.ClearHighBitCommand


class byte_bitwise_not(IndexRangeAction):
    """Invert every bit in each byte"""
    cmd = commands.BitwiseNotCommand


class byte_shift_left(IndexRangeAction):
    """Shift bits left (multiply by 2), inserting zeros in the low bit"""
    cmd = commands.LeftShiftCommand


class byte_shift_right(IndexRangeAction):
    """Shift bits right (divide by 2), inserting zeros in the high bit"""
    cmd = commands.RightShiftCommand


class byte_rotate_left(IndexRangeAction):
    """Shift bits left where the high bit wraps around to become the new low bit"""
    cmd = commands.LeftRotateCommand


class byte_rotate_right(IndexRangeAction):
    """Shift bits right where the low bit wraps around to become the new high bit"""
    cmd = commands.RightRotateCommand


class byte_reverse_bits(IndexRangeAction):
    """Reverse the bit pattern of each byte; e.g. $c0 or 11000000 in binary becomes 00000011 in binary, $03 in hex"""
    cmd = commands.ReverseBitsCommand


class byte_random(IndexRangeAction):
    """Generate random bytes to replace the selected data"""
    cmd = commands.RandomBytesCommand


class IndexRangeValueAction(IndexRangeAction):
    prompt = "Enter byte value: (default hex, prefix with # for decimal, %% for binary)"

    def calc_name(self, action_key):
        return self.cmd.ui_name + "..."

    def show_dialog(self):
        value = prompt_for_hex(self.viewer.control, self.prompt, self.cmd.ui_name)
        if value is not None:
            ranges = self.viewer.control.get_selected_ranges()
            cmd = self.cmd(self.editor.segment, ranges, value)
            self.editor.process_command(cmd)

    def perform(self, action_key):
        wx.CallAfter(self.show_dialog)

    def _update_enabled(self, ui_state):
        # Setting a range of values don't make sense for a single byte
        # location, so require it to have an actual range
        return self.viewer.has_editable_bytes and self.linked_base.has_selection


class byte_set_value(IndexRangeValueAction):
    """Prompts the user and sets the data to the specified value"""
    cmd = commands.SetValueCommand


class byte_or_with_value(IndexRangeValueAction):
    """Logical OR the selected data with the user specified value"""
    cmd = commands.OrWithCommand


class byte_and_with_value(IndexRangeValueAction):
    """Logical AND the selected data with the user specified value"""
    cmd = commands.AndWithCommand


class byte_xor_with_value(IndexRangeValueAction):
    """Logical XOR the selected data with the user specified value"""
    cmd = commands.XorWithCommand


class SliceValueAction(IndexRangeValueAction):
    prompt = "Enter byte value: (default hex, prefix with # for decimal, %% for binary)"

    def show_dialog(self):
        slice_obj = prompt_for_slice(self.viewer.control, self.prompt, self.cmd.ui_name)
        if slice_obj is not None:
            cmd = self.cmd(self.linked_base.segment, self.linked_base.carets.selected_ranges, slice_obj)
            self.editor.process_command(cmd)


class byte_ramp_up(SliceValueAction):
    """Starting with the user specified value at the first selected byte, loops
    over each byte in the selection and adds one to the value of the previous
    byte. At $ff, it wraps around to $00.
    """
    cmd = commands.RampUpCommand


class byte_ramp_down(SliceValueAction):
    """Starting with the user specified value at the first selected byte, loops
    over each byte in the selection and subtracts one from the value of the
    previous byte. At $00, it wraps around to $ff.
    """
    cmd = commands.RampDownCommand


class byte_add_value(IndexRangeValueAction):
    """Adds the user specified value to the data, performing a logical AND with
    $ff if necessary to keep all values in the 8-bit range."""
    cmd = commands.AddValueCommand


class byte_subtract_value(IndexRangeValueAction):
    """Subtracts the user specified value from the data (AND with $ff if
    necessary). Note the difference between this and `Subtract From`_"""
    cmd = commands.SubtractValueCommand


class byte_subtract_from(IndexRangeValueAction):
    """Subtracts the data from the user specified value (AND with $ff if
    necessary). Note the difference between this and `Subtract`_"""
    cmd = commands.SubtractFromCommand


class byte_multiply_by(IndexRangeValueAction):
    """Multiply the data from the user specified value (AND with $ff if
    necessary)."""
    cmd = commands.MultiplyCommand


class byte_divide_by(IndexRangeValueAction):
    """Divides the data by the user specified value by the data, ignoring the
    remainder. Note the difference between this and `Divide From`_"""
    cmd = commands.DivideByCommand


class byte_divide_from(IndexRangeValueAction):
    """Divides the data from the user specified value (that is to say: dividing
    the user specified value by the data), ignoring the remainder. Note the
    difference between this and `Divide By`_"""
    cmd = commands.DivideFromCommand


class byte_reverse_selection(IndexRangeAction):
    """Reverses the order of bytes in the selection"""
    cmd = commands.ReverseSelectionCommand


class byte_reverse_group(IndexRangeValueAction):
    prompt = "Enter number of bytes in each group: (default hex, prefix with # for decimal, %% for binary)"
    cmd = commands.ReverseGroupCommand



from __future__ import division
import cython
import numpy as np
cimport numpy as np

# from atrcopy; doesn't seem to be a way to import Python constants
# (although, to be fair, nothing in Python can be assumed to be a constant!)
cdef enum:
    user_bit_mask = 0x07
    data_style = 0x1
    not_user_bit_mask = 0xff ^ user_bit_mask
    diff_bit_mask = 0x10
    match_bit_mask = 0x20
    comment_bit_mask = 0x40
    selected_bit_mask = 0x80
    ignore_mask = not_user_bit_mask & (0xff ^ diff_bit_mask)
    invalid_style = 0xff


@cython.boundscheck(False)
@cython.wraparound(False)
def calc_rgb_from_color_indexes_naive(np.ndarray[np.uint8_t, ndim=2] color_indexes, np.ndarray[np.uint8_t, ndim=2] style_per_pixel, colors, empty_color_tuple):
    cdef int h = color_indexes.shape[0]
    cdef int w = color_indexes.shape[1]

    cdef np.uint8_t[:] color_indexes_flat = color_indexes.reshape(-1)
    cdef np.uint8_t[:] style_per_pixel_flat = style_per_pixel.reshape(-1)
    cdef np.ndarray[np.uint8_t, ndim=2] flat_image = np.empty([h * w, 3], dtype=np.uint8)

    cdef np.ndarray[np.uint8_t, ndim=2] color_registers = colors[0]
    cdef np.ndarray[np.uint8_t, ndim=2] h_colors = colors[1]
    cdef np.ndarray[np.uint8_t, ndim=2] m_colors = colors[2]
    cdef np.ndarray[np.uint8_t, ndim=2] c_colors = colors[3]
    cdef np.ndarray[np.uint8_t, ndim=2] d_colors = colors[4]
    cdef np.ndarray[np.uint8_t, ndim=1] empty_color = np.asarray(empty_color_tuple, dtype=np.uint8)

    cdef np.uint8_t color_index, style
    cdef int i
    for i in range(len(color_indexes_flat)):
        color_index = color_indexes_flat[i]
        style = style_per_pixel_flat[i]
        if style == invalid_style:
            flat_image[i] = empty_color
        elif style & ignore_mask == 0:
            flat_image[i] = color_registers[color_index]
        elif style & selected_bit_mask:
            flat_image[i] = h_colors[color_index]
        elif (style & user_bit_mask) > 0:
            flat_image[i] = d_colors[color_index]
        elif style & comment_bit_mask:
            flat_image[i] = c_colors[color_index]
        elif style & match_bit_mask:
            flat_image[i] = h_colors[color_index]
        else:  # not any of the above, which shouldn't happen
            flat_image[i] = (0xff, 0, 0xee)
    return flat_image.reshape((h, w, 3))


@cython.boundscheck(False)
@cython.wraparound(False)
def calc_rgb_from_color_indexes_better(np.ndarray[np.uint8_t, ndim=2] color_indexes, np.ndarray[np.uint8_t, ndim=2] style_per_pixel, colors, empty_color_tuple):
    cdef int h = color_indexes.shape[0]
    cdef int w = color_indexes.shape[1]

    cdef np.uint8_t[:] color_indexes_flat = color_indexes.reshape(-1)
    cdef np.uint8_t[:] style_per_pixel_flat = style_per_pixel.reshape(-1)
    cdef np.ndarray[np.uint8_t, ndim=2] flat_image = np.empty([h * w, 3], dtype=np.uint8)

    cdef np.ndarray[np.uint8_t, ndim=2] color_registers = colors[0]
    cdef np.ndarray[np.uint8_t, ndim=2] h_colors = colors[1]
    cdef np.ndarray[np.uint8_t, ndim=2] m_colors = colors[2]
    cdef np.ndarray[np.uint8_t, ndim=2] c_colors = colors[3]
    cdef np.ndarray[np.uint8_t, ndim=2] d_colors = colors[4]
    cdef np.ndarray[np.uint8_t, ndim=1] empty_color = np.asarray(empty_color_tuple, dtype=np.uint8)

    cdef np.uint8_t color_index, style, r, g, b
    cdef int i
    for i in range(len(color_indexes_flat)):
        color_index = color_indexes_flat[i]
        style = style_per_pixel_flat[i]
        if style == invalid_style:
            r = empty_color[0]
            g = empty_color[1]
            b = empty_color[2]
        elif style & ignore_mask == 0:
            r = color_registers[color_index, 0]
            g = color_registers[color_index, 1]
            b = color_registers[color_index, 2]
        elif style & selected_bit_mask:
            r = h_colors[color_index, 0]
            g = h_colors[color_index, 1]
            b = h_colors[color_index, 2]
        elif (style & user_bit_mask) > 0:
            r = d_colors[color_index, 0]
            g = d_colors[color_index, 1]
            b = d_colors[color_index, 2]
        elif style & comment_bit_mask:
            r = c_colors[color_index, 0]
            g = c_colors[color_index, 1]
            b = c_colors[color_index, 2]
        elif style & match_bit_mask:
            r = h_colors[color_index, 0]
            g = h_colors[color_index, 1]
            b = h_colors[color_index, 2]
        else:  # not any of the above, which shouldn't happen
            r = 0xff
            g = 0
            b = 0xee

        flat_image[i,0] = r
        flat_image[i,1] = g
        flat_image[i,2] = b
    return flat_image.reshape((h, w, 3))


@cython.boundscheck(False)
@cython.wraparound(False)
def calc_rgb_from_color_indexes_fast(np.ndarray[np.uint8_t, ndim=2] color_indexes, np.ndarray[np.uint8_t, ndim=2] style_per_pixel, colors, empty_color_tuple):
    cdef int h = color_indexes.shape[0]
    cdef int w = color_indexes.shape[1]

    cdef np.uint8_t[:] color_indexes_flat = color_indexes.reshape(-1)
    cdef np.uint8_t[:] style_per_pixel_flat = style_per_pixel.reshape(-1)
    cdef np.ndarray[np.uint8_t, ndim=2] flat_image = np.empty([h * w, 3], dtype=np.uint8)
    cdef np.uint8_t[:,:] flat_image_fast = flat_image

    cdef np.ndarray[np.uint8_t, ndim=2] color_registers = colors[0]
    cdef np.uint8_t[:,:] color_registers_fast = color_registers
    cdef np.ndarray[np.uint8_t, ndim=2] h_colors = colors[1]
    cdef np.uint8_t[:,:] h_colors_fast = h_colors
    cdef np.ndarray[np.uint8_t, ndim=2] m_colors = colors[2]
    cdef np.uint8_t[:,:] m_colors_fast = m_colors
    cdef np.ndarray[np.uint8_t, ndim=2] c_colors = colors[3]
    cdef np.uint8_t[:,:] c_colors_fast = c_colors
    cdef np.ndarray[np.uint8_t, ndim=2] d_colors = colors[4]
    cdef np.uint8_t[:,:] d_colors_fast = d_colors
    cdef np.uint8_t empty_r, empty_g, empty_b
    empty_r = empty_color_tuple[0]
    empty_g = empty_color_tuple[1]
    empty_b = empty_color_tuple[2]

    cdef np.uint8_t color_index, style, r, g, b
    cdef int i
    for i in range(len(color_indexes_flat)):
        color_index = color_indexes_flat[i]
        style = style_per_pixel_flat[i]
        if style == invalid_style:
            r = empty_r
            g = empty_g
            b = empty_b
        elif style & ignore_mask == 0:
            r = color_registers_fast[color_index, 0]
            g = color_registers_fast[color_index, 1]
            b = color_registers_fast[color_index, 2]
        elif style & selected_bit_mask:
            r = h_colors_fast[color_index, 0]
            g = h_colors_fast[color_index, 1]
            b = h_colors_fast[color_index, 2]
        elif (style & user_bit_mask) > 0:
            r = d_colors_fast[color_index, 0]
            g = d_colors_fast[color_index, 1]
            b = d_colors_fast[color_index, 2]
        elif style & comment_bit_mask:
            r = c_colors_fast[color_index, 0]
            g = c_colors_fast[color_index, 1]
            b = c_colors_fast[color_index, 2]
        elif style & match_bit_mask:
            r = h_colors_fast[color_index, 0]
            g = h_colors_fast[color_index, 1]
            b = h_colors_fast[color_index, 2]
        else:  # not any of the above, which shouldn't happen
            r = 0xff
            g = 0
            b = 0xee

        flat_image_fast[i,0] = r
        flat_image_fast[i,1] = g
        flat_image_fast[i,2] = b
    return flat_image.reshape((h, w, 3))


@cython.boundscheck(False)
@cython.wraparound(False)
def calc_rgb_from_color_indexes_fast_optimize(np.ndarray[np.uint8_t, ndim=2] color_indexes, np.ndarray[np.uint8_t, ndim=2] style_per_pixel, colors, empty_color_tuple):
    cdef int h = color_indexes.shape[0]
    cdef int w = color_indexes.shape[1]

    cdef np.uint8_t[:] color_indexes_flat = color_indexes.reshape(-1)
    cdef np.uint8_t[:] style_per_pixel_flat = style_per_pixel.reshape(-1)
    cdef np.ndarray[np.uint8_t, ndim=2] flat_image = np.empty([h * w, 3], dtype=np.uint8)
    cdef np.uint8_t[:,:] flat_image_fast = flat_image

    cdef np.ndarray[np.uint8_t, ndim=2] color_registers = colors[0]
    cdef np.uint8_t[::1] color_registers_fast_temp = color_registers
    cdef np.uint8_t[:] color_registers_fast = color_registers_fast_temp
    cdef np.ndarray[np.uint8_t, ndim=2] h_colors = colors[1]
    cdef np.uint8_t[:,:] h_colors_fast = h_colors
    cdef np.ndarray[np.uint8_t, ndim=2] m_colors = colors[2]
    cdef np.uint8_t[:,:] m_colors_fast = m_colors
    cdef np.ndarray[np.uint8_t, ndim=2] c_colors = colors[3]
    cdef np.uint8_t[:,:] c_colors_fast = c_colors
    cdef np.ndarray[np.uint8_t, ndim=2] d_colors = colors[4]
    cdef np.uint8_t[:,:] d_colors_fast = d_colors
    cdef np.uint8_t empty_r, empty_g, empty_b
    empty_r = empty_color_tuple[0]
    empty_g = empty_color_tuple[1]
    empty_b = empty_color_tuple[2]

    cdef np.uint8_t style, r, g, b
    cdef int i, color_index
    for i in range(len(color_indexes_flat)):
        color_index = color_indexes_flat[i] * 3
        style = style_per_pixel_flat[i]
        if style == invalid_style:
            r = empty_r
            g = empty_g
            b = empty_b
        elif style & ignore_mask == 0:
            r = color_registers_fast[color_index]
            color_index += 1
            g = color_registers_fast[color_index]
            color_index += 1
            b = color_registers_fast[color_index]
        elif style & selected_bit_mask:
            r = h_colors_fast[color_index, 0]
            g = h_colors_fast[color_index, 1]
            b = h_colors_fast[color_index, 2]
        elif (style & user_bit_mask) > 0:
            r = d_colors_fast[color_index, 0]
            g = d_colors_fast[color_index, 1]
            b = d_colors_fast[color_index, 2]
        elif style & comment_bit_mask:
            r = c_colors_fast[color_index, 0]
            g = c_colors_fast[color_index, 1]
            b = c_colors_fast[color_index, 2]
        elif style & match_bit_mask:
            r = h_colors_fast[color_index, 0]
            g = h_colors_fast[color_index, 1]
            b = h_colors_fast[color_index, 2]
        else:  # not any of the above, which shouldn't happen
            r = 0xff
            g = 0
            b = 0xee

        flat_image_fast[i,0] = r
        flat_image_fast[i,1] = g
        flat_image_fast[i,2] = b
    return flat_image.reshape((h, w, 3))


# The best seems to be better

calc_rgb_from_color_indexes = calc_rgb_from_color_indexes_better



import ctypes
import time

import numpy as np
np.set_printoptions(formatter={'int':hex})

try:
    import wx
except ImportError:
    wx = None

from ...utils import apple2util as a2

from ..generic6502 import lib6502, Generic6502
from ..generic6502 import dtypes as d

import logging
logging.basicConfig(level=logging.WARNING)
log = logging.getLogger(__name__)


debug_frames = False


def a2_color_map():
    rmap = np.zeros(256, dtype=np.uint8)
    gmap = np.zeros(256, dtype=np.uint8)
    bmap = np.zeros(256, dtype=np.uint8)

    rmap[1] = 255
    gmap[1] = 255
    bmap[1] = 255

    return rmap, gmap, bmap


class Crabapple(Generic6502):
    cpu = "6502"
    name = "crabapple"
    ui_name = "Crabapple 2+"

    width = 560
    height = 384

    supported_filesystems = []
    supported_binaries = ["Apple DOS 3.3 Object File"]

    def configure_emulator_defaults(self):
        lib6502.set_a2_emulation_mode(1)

    def generate_save_state_memory_blocks(self):
        cpu_offset = self.state_start_offset
        memory_offset = cpu_offset + d.CPU_DTYPE.itemsize
        video_offset = cpu_offset + d.STATESAV_MAX_SIZE
        segments = [
            (cpu_offset, d.CPU_DTYPE.itemsize, 0, "CPU Status"),
            (memory_offset, d.MAIN_MEMORY_SIZE, 0, "Main Memory"),
            (memory_offset + 0x400, 0x400, 0x400, "Text/Lo-res Page 1"),
            (memory_offset + 0x800, 0x400, 0x800, "Text/Lo-res Page 2"),
            (memory_offset + 0x2000, 0x2000, 0x2000, "Hi-res Page 1"),
            (memory_offset + 0x4000, 0x2000, 0x4000, "Hi-res Page 2"),
            (video_offset, d.VIDEO_SIZE, 0, "Video RAM"),
            (video_offset + d.VIDEO_SIZE, d.VIDEO_HEIGHT, 0, "Scan Line Type"),
        ]
        self.save_state_memory_blocks.extend(segments)

    def compute_color_map(self):
        self.rmap, self.gmap, self.bmap = a2_color_map()

    def get_color_indexed_screen(self, frame_number=-1):
        if frame_number < 0:
            output = self.output
        else:
            _, output = self.get_history(frame_number)
        doubled = np.empty((384, 40), dtype=np.uint8)
        source = output['video'].reshape((192, 40))
        doubled[::2,:] = source
        doubled[1::2,:] = source
        raw = np.empty(self.height * self.width, dtype=np.uint8)
        a2.to_560_bw_pixels(doubled, raw)
        #print "get_raw_screen", frame_number, raw
        return raw.reshape((self.height, self.width))

    def get_frame_rgb(self, frame_number=-1):
        raw = self.get_color_indexed_screen(frame_number)
        self.screen_rgb[:,:,0] = self.rmap[raw]
        self.screen_rgb[:,:,1] = self.gmap[raw]
        self.screen_rgb[:,:,2] = self.bmap[raw]
        return self.screen_rgb

    if wx is not None:
        def process_key_down(self, evt, keycode):
            log.debug("key down! key=%s mod=%s" % (evt.GetKeyCode(), evt.GetModifiers()))
            key = evt.GetKeyCode()
            if key > 96 and key < 123:
                key -= 32
            elif key == wx.WXK_UP:
                key = 0x8b
            elif key == wx.WXK_DOWN:
                key = 0x8a
            elif key == wx.WXK_LEFT:
                key = 0x88
            elif key == wx.WXK_RIGHT:
                key = 0x95
            key |= 0x80
            self.send_char(key)



import numpy as np

import wx
import wx.lib.buttons as buttons
import wx.lib.sized_controls as sc
import wx.lib.colourselect as csel
import wx.lib.colourchooser.canvas as canvas

from atrip.machines import atari8bit


#704  2C0  PCOLR0
#705  2C1  PCOLR1
#706  2C2  PCOLR2
#707  2C3  PCOLR3
#708  2C4  COLOR0
#709  2C5  COLOR1
#710  2C6  COLOR2
#711  2C7  COLOR3
#712  2C8  COLOR4 (Background/Border)

default_color_register_names = ["%d (%x) %s" % (i + 704, i + 704, n) for i, n in enumerate([
    "Player 0",
    "Player 1",
    "Player 2",
    "Player 3",
    "Playfield 0",
    "Playfield 1",
    "Playfield 2",
    "Playfield 3",
    "Background",
    ])]


class ColorListBox(wx.VListBox):
    def calc_sizes(self):
        self.max_w = 0
        n = 0
        while True:
            try:
                label = self.get_label(n)
            except IndexError:
                break
            w, h = self.GetTextExtent(label)
            if w > self.max_w:
                self.max_w = w
            n += 1

    def get_label(self, n):
        names = self.GetParent().color_register_names
        return names[n]

    # This method must be overridden.  When called it should draw the
    # n'th item on the dc within the rect.  How it is drawn, and what
    # is drawn is entirely up to you.
    def OnDrawItem(self, dc, rect, n):
        if self.GetSelection() == n:
            c = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT)
        else:
            c = self.GetForegroundColour()
        dc.SetFont(self.GetFont())
        dc.SetTextForeground(c)
        label = self.get_label(n)
        dc.DrawLabel(label, rect, wx.ALIGN_LEFT | wx.ALIGN_CENTER_VERTICAL)
        color = self.GetParent().get_color(n)
        dc.SetBrush(wx.Brush(color))
        rect.Offset(self.max_w, 0)
        rect.Deflate(2, 2)
        dc.DrawRectangle(rect)

    # This method must be overridden.  It should return the height
    # required to draw the n'th item.
    def OnMeasureItem(self, n):
        label = self.get_label(n)
        w, h = self.GetTextExtent(label)
        return h + 5


class AnticPalette(canvas.Canvas):
    """Large image selection using Antic colors
    """

    HORIZONTAL_STEP = 24
    VERTICAL_STEP = 24
    BORDER = 4
    HIGHLIGHT_WIDTH = 5

    def __init__(self, parent, color_prefs):
        """Creates a palette object."""
        # Load the pre-generated palette XPM

        # Leaving this in causes warning messages in some cases.
        # It is the responsibility of the app to init the image
        # handlers, IAW RD
        #wx.InitAllImageHandlers()
        self.color_prefs = color_prefs
        self.palette = self.init_palette()
        size = self.palette.GetSize()
        canvas.Canvas.__init__(self, parent, -1, style=wx.SIMPLE_BORDER, forceClientSize=size)

    def init_palette(self):
        w = self.HORIZONTAL_STEP * 16 + 2 * self.BORDER
        h = self.VERTICAL_STEP * 16 + 2 * self.BORDER
        array = np.empty((h, w, 3), dtype=np.uint8)
        array[:,:] = self.color_prefs.empty_background_color[:3]
        for high in range(0, 256, 16):
            y = self.BORDER + (high // 16) * self.VERTICAL_STEP
            for low in range(16):
                x = self.BORDER + low * self.HORIZONTAL_STEP
                c = atari8bit.gtia_ntsc_to_rgb(high + low)
                array[y:y+self.VERTICAL_STEP,x:x+self.HORIZONTAL_STEP,:] = c
        width = array.shape[1]
        height = array.shape[0]
        image = wx.Image(width, height)
        image.SetData(array.tobytes())
        bmp = wx.Bitmap(image)
        return bmp

    def DrawBuffer(self):
        """Draws the palette XPM into the memory buffer."""
        #self.GeneratePaletteBMP ("foo.bmp")
        self.buffer.DrawBitmap(self.palette, 0, 0, 0)

    def highlight_xy(self, x, y):
        """Highlights an area of the palette with a rectangle around
        the coordinate point"""
        low = (x - self.BORDER) // self.HORIZONTAL_STEP
        high = (y - self.BORDER) // self.VERTICAL_STEP
        return self.highlight(high, low)

    def highlight_color(self, c):
        high, low = divmod(c, 16)
        return self.highlight(high, low)

    def highlight(self, high, low):
        colour = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT)
        self.buffer.SetPen(wx.Pen(colour, self.HIGHLIGHT_WIDTH, wx.SOLID))
        self.buffer.SetBrush(wx.Brush(colour, wx.TRANSPARENT))
        low = 15 if low > 15 else low
        low = 0 if low < 0 else low
        high = 15 if high > 15 else high
        high = 0 if high < 0 else high
        x = self.BORDER + low * self.HORIZONTAL_STEP
        y = self.BORDER + high * self.VERTICAL_STEP
        self.buffer.DrawRectangle(x, y, self.HORIZONTAL_STEP, self.VERTICAL_STEP)
        self.Refresh()
        return high * 16 + low


class AnticColorDialog(wx.Dialog):
    def __init__(self, parent, antic_color_registers, color_prefs, color_register_names=None):
        wx.Dialog.__init__(self, parent, -1, "Choose ANTIC Colors")
        self.num_cols = 17
        if color_register_names is None:
            self.color_register_names = default_color_register_names
        else:
            self.color_register_names = color_register_names

        sizer = wx.BoxSizer(wx.VERTICAL)

        self.name = wx.StaticText(self, -1, "Editing Color")
        sizer.Add(self.name, 0, wx.ALIGN_CENTRE|wx.ALL, 5)

        box = wx.BoxSizer(wx.HORIZONTAL)
        self.color_registers = ColorListBox(self, -1, style=wx.VSCROLL, size=(200, -1))
        box.Add(self.color_registers, 0, wx.EXPAND|wx.ALL, 4)
        self.palette = AnticPalette(self, color_prefs)
        box.Add(self.palette, 1, wx.ALL, 4)
        sizer.Add(box, 1, wx.GROW|wx.ALL, 5)

        btn_sizer = self.CreateStdDialogButtonSizer(wx.OK | wx.CANCEL)
        sizer.Add(btn_sizer, 0, wx.GROW|wx.ALL, 12)

        self.Bind(wx.EVT_CLOSE, self.on_close)
        self.Bind(wx.EVT_BUTTON, self.on_button)
        self.Bind(wx.EVT_LISTBOX, self.on_set_register)
        self.mouse_down = False
        self.palette.Bind(wx.EVT_LEFT_DOWN, self.onPaletteDown)
        self.palette.Bind(wx.EVT_LEFT_UP, self.onPaletteUp)
        self.palette.Bind(wx.EVT_MOTION, self.onPaletteMotion)

        self.init_colors(antic_color_registers)

        # Need SetSizeHints to force the window to fit the size of the sizer. From
        # http://wxpython.org/Phoenix/docs/html/sizers_overview.html#sizers-overview
        sizer.SetSizeHints(self)
        self.SetSizer(sizer)
        self.Layout()
        self.Fit()

    def init_colors(self, antic_color_registers):
        self.color_registers.calc_sizes()
        c = list(antic_color_registers)
        if len(c) == 5:
            self.colors = list([0, 0, 0, 0])
            self.colors.extend(list(c))
        else:
            self.colors = c
        self.color_registers.SetItemCount(9)
        self.current_register = 4
        self.color_registers.SetSelection(self.current_register)

    def get_color(self, register):
        return atari8bit.gtia_ntsc_to_rgb(self.colors[register])

    def on_button(self, event):
        if event.GetId() == wx.ID_OK:
            self.EndModal(wx.ID_OK)
            event.Skip()
        else:
            self.EndModal(wx.ID_CANCEL)
            event.Skip()

    def on_close(self, event):
        self.EndModal(wx.ID_CANCEL)
        event.Skip()

    def on_set_register(self, event):
        self.current_register = event.GetSelection()
        color = self.colors[self.current_register]
        self.palette.ReDraw()
        self.palette.highlight_color(color)

    def on_set_color(self, event):
        id = event.GetId()
        b, c = self.toggles[id]
        self.colors[self.current_register] = c
        self.color_registers.Refresh()

    def onPaletteDown(self, event):
        """Stores state that the mouse has been pressed and updates
        the selected colour values."""
        self.mouse_down = True
        self.palette.ReDraw()
        self.doPaletteClick(event.X, event.Y)

    def onPaletteUp(self, event):
        """Stores state that the mouse is no longer depressed."""
        self.mouse_down = False

    def onPaletteMotion(self, event):
        """Updates the colour values during mouse motion while the
        mouse button is depressed."""
        if self.mouse_down:
            self.doPaletteClick(event.X, event.Y)

    def doPaletteClick(self, m_x, m_y):
        """Updates the colour values based on the mouse location
        over the palette."""
        # Highlight a fresh selected area
        self.palette.ReDraw()
        c = self.palette.highlight_xy(m_x, m_y)
        self.colors[self.current_register] = c
        self.color_registers.Refresh()


if __name__ == "__main__":
    """
    simple test for the dialog
    """
    from omnivore.arch.machine import Machine

    a = wx.App(False)
    import wx.lib.inspection
    wx.lib.inspection.InspectionTool().Show()
    w = wx.Frame(None)
    attr = w.GetDefaultAttributes()
    antic_color_registers = [40, 202, 148, 70, 0]
    Machine.empty_color = attr.colBg.Get(False)
    d = AnticColorDialog(w, antic_color_registers)
    d.Show()
    a.MainLoop()



import os
import sys

import numpy as np

import wx

from sawx.ui import compactgrid as cg
from ..editors.linked_base import VirtualTableLinkedBase

from ..ui.segment_grid import SegmentGridControl, SegmentTable

from ..viewer import SegmentViewer

from ..commands.disasm import MiniAssemblerCommand
from ..commands.comment import SetCommentCommand
from atrip.disassembler import get_miniasm
from ..utils import searchutil

import logging
log = logging.getLogger(__name__)


class DisassemblyTable(SegmentTable):
    column_labels = ["^Disassembly"]
    column_sizes = [48]

    def __init__(self, linked_base):
        SegmentTable.__init__(self, linked_base, len(self.column_labels), False)

        self.max_num_entries = 80000
        self.rebuild()

    def calc_num_rows(self):
        try:
            return len(self.current)
        except AttributeError:
            return 0

    def get_index_range(self, row, cell):
        """Get the byte offset from start of file given row, col
        position.
        """
        e = self.current.entries
        index = e[row]['pc'] - self.current.origin
        return index, index + e[row]['num_bytes']

    def get_index_of_row(self, row):
        index, _ = self.get_index_range(row, 0)
        return index

    def get_start_end_index_of_row(self, row):
        index1, index2 = self.get_index_range(row, 0)
        return index1, index2

    def index_to_row_col(self, index):
        return self.current.index_to_row[index], 0

    def get_label_at_index(self, index):
        row, _ = self.index_to_row_col(index)
        return str(self.current.entries[row]['pc'])

    def get_row_label_text(self, start_line, num_lines, step=1):
        last_line = min(start_line + num_lines, self.num_rows)
        entries = self.current.entries
        for line in range(start_line, last_line, step):
            yield "%04x" % (entries[line]['pc'])

    def get_value_style(self, row, col):
        index, _ = self.get_index_range(row, col)
        style = 0
        if self.is_index_valid(index):
            s = self.linked_base.segment
            p = self.current
            e = p.entries
            addr = e[row]['pc']
            has_label = p.jmp_targets[addr]
            if has_label:
                text = "L%04x " % addr
            else:
                text = "      "

            t = self.parsed
            if t is not None:
                text += t[row - t.start_index].decode('utf-8')

            comments = []
            for i in range(index, index + e[row]['num_bytes']):
                c = s.get_comment_at(i)
                if c:
                    comments.append(c)
            comment = " ".join([str(c) for c in comments])
            if comment:
                if len(text) < 20:
                    text += " " * (20 - len(text))
                text += " ; " + comment

            for i in range(index, index + e[row]['num_bytes']):
                style |= s.style[i]
        else:
            text = ""
        # print(f"get_value_style: {row},{col} = {index} {self.last_valid_index} {self.is_index_valid(index)} ; {text}, {style}, {self.linked_base.segment}")
        return text, style

    def prepare_for_drawing(self, start_row, visible_rows, start_cell, visible_cells):
        self.parsed = self.current.stringify(start_row, visible_rows, self.linked_base.document.labels)

    def search(self, search_bytes, match_case=False):
        return self.current.search(search_bytes, match_case, self.linked_base.document.labels)

    def rebuild(self):
        segment = self.linked_base.segment
        self.current = self.linked_base.document.disassembler.parse(segment, self.max_num_entries)
        self.parsed = None
        self.init_boundaries()


class DisassemblyControl(SegmentGridControl):
    default_table_cls = DisassemblyTable

    def calc_default_table(self, linked_base):
        return self.default_table_cls(linked_base)

    def calc_line_renderer(self):
        return cg.VirtualTableLineRenderer(self, 2, widths=self.default_table_cls.column_sizes, col_labels=self.default_table_cls.column_labels)

    def set_viewer_defaults(self):
        self.want_col_header = False
        super().set_viewer_defaults()

    def recalc_view(self):
        self.table.rebuild()
        super().recalc_view()

    def calc_ranges_for_edit(self):
        table = self.table
        ranges = []
        for c in self.caret_handler.carets:
            r = c.range_including_caret
            start, _ = table.get_index_range(*r[0])
            _, end = table.get_index_range(*r[1])
            row1, _ = table.index_to_row_col(start)
            row2, _ = table.index_to_row_col(end - 1)
            if row1 == row2:
                # single line selected, so pass a singly byte so moving from a
                # multi-byte op (e.g. JSR) to fewer bytes doesn't result in
                # more than one op
                ranges.append((start, start + 1))
            else:
                # multiple lines passes entire range, MiniAssemblerCommand
                # handdles possible overlaps
                ranges.append((start, end))
        return ranges

    def calc_edit_command(self, val):
        if val.startswith(";"):
            ranges = self.caret_handler.calc_ranges(self.segment_viewer.segment)
            val = val[1:].strip()
            cmd = SetCommentCommand(self.segment_viewer.segment, ranges, val)
        else:
            cmd = MiniAssemblerCommand(self.segment_viewer.segment, self.segment_viewer.document.cpu, self.caret_handler, val, advance=True)
        return cmd

    def advance_caret_position(self, evt, flags):
        self.caret_handler.move_carets_vertically(self.table, 1)

    def verify_keycode_can_start_edit(self, c):
        if c == 59:  # semicolon
            return True
        miniasm = get_miniasm(self.segment_viewer.document.cpu)
        return miniasm.can_start_edit(c)


# Disassembly searcher uses the __call__ method to return the object because it
# needs access to the particular viewer's CPU type and disassembly table.
# Normal searchers just use the segment's raw data and returns itself in the
# constructor.
class DisassemblySearcher(searchutil.BaseSearcher):
    def __init__(self, viewer):
        self.search_text = None
        self.matches = []
        self.table = viewer.table
        self.ui_name = viewer.document.cpu

    def __call__(self, editor, search_text, search_copy):
        self.search_text = self.get_search_text(search_text)
        if len(self.search_text) > 0:
            self.matches = self.get_matches(editor)
        else:
            self.matches = []
        return self

    def __str__(self):
        return f"{self.ui_name} matches: {self.matches}"

    def get_matches(self, editor):
        match_case = editor.last_search_settings.get('match_case', False)
        search_bytes = self.search_text
        if not match_case:
            search_bytes = search_bytes.lower()
        matches = self.table.search(search_bytes, match_case)
        log.debug("instruction matches: %s" % str(matches))
        return matches


class DisassemblyViewer(SegmentViewer):
    name = "disasm"

    ui_name = "Static Disassembly"

    control_cls = DisassemblyControl

    # initialization

    # properties

    @property
    def table(self):
        return self.control.table

    @property
    def searchers(self):
        # Replace the normal viewer searcher class attribute with a searcher
        # that is custom to the cpu shown in this viewer
        return [DisassemblySearcher(self)]

    def set_event_handlers(self):
        self.document.cpu_changed_event += self.on_cpu_changed
        super().set_event_handlers()

    def on_cpu_changed(self, evt):
        self.recalc_view()
        self.refresh_view(True)

    def on_update_table_for_value_change(self, evt):
        self.table.rebuild()

    def on_update_table_for_style_change(self, evt):
        self.table.rebuild()

    def recalc_data_model(self):
        self.control.recalc_view()



import time

import wx

from sawx import get_image_path
from sawx.utils.command import DisplayFlags
from sawx.framework.caret import Caret, SelectionHandler

import logging
log = logging.getLogger(__name__)


class MouseEventMixin(SelectionHandler):
    hand_cursor = None
    hand_closed_cursor = None

    def __init__(self, caret_handler, default_mouse_mode_cls):
        self.caret_handler = caret_handler
        self.multi_select_mode = False
        self.select_extend_mode = False
        self.mouse_drag_started = False
        self.is_editing_in_cell = False
        self.is_mousing_while_editing = False
        self.pending_select_awaiting_drag = None
        self.caret_with_selection = None
        self.source = None
        if self.__class__.hand_cursor is None:
            p = get_image_path("icons/hand.ico")
            self.hand_cursor = wx.Cursor(p, wx.BITMAP_TYPE_ICO, 16, 16)
            p = get_image_path("icons/hand_closed.ico")
            self.hand_closed_cursor = wx.Cursor(p, wx.BITMAP_TYPE_ICO, 16, 16)
        self.forced_cursor = None
        self.batch = None

        self.default_mouse_mode_cls = default_mouse_mode_cls
        self.default_mouse_mode = self.default_mouse_mode_cls(self)
        self.mouse_mode = self.default_mouse_mode  # can't call set_mouse_mode yet because control hasn't been initialized

    def map_mouse_events(self, source):
        self.source = source
        source.Bind(wx.EVT_MOUSEWHEEL, self.on_mouse_wheel)
        source.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)
        source.Bind(wx.EVT_MOTION, self.on_motion)
        source.Bind(wx.EVT_LEFT_UP, self.on_left_up)
        source.Bind(wx.EVT_LEFT_DCLICK, self.on_left_dclick)
        source.Bind(wx.EVT_RIGHT_DOWN, self.on_popup)
        source.Bind(wx.EVT_MOUSEWHEEL, self.on_mouse_wheel)
        source.Bind(wx.EVT_SET_FOCUS, self.on_focus)
        source.Bind(wx.EVT_KILL_FOCUS, self.on_focus_lost)
        source.Bind(wx.EVT_ENTER_WINDOW, self.on_mouse_enter)
        source.Bind(wx.EVT_LEAVE_WINDOW, self.on_mouse_leave)

    def set_mouse_mode(self, mode=None):
        self.release_mouse()
        if mode is None:
            mode = self.default_mouse_mode_cls
        if self.mouse_mode.__class__ != mode:
            log.debug("set_mouse_mode: %s" % mode)
            self.mouse_mode = mode(self)
        else:
            log.debug("mouse mode already %s" % mode)

    def capture_mouse(self):
        if not self.source.HasFocus():
            self.source.SetFocus()
        self.source.CaptureMouse()

    def has_capture(self):
        return self.source.HasCapture()

    def release_mouse(self):
        self.mouse_is_down = False
        self.selection_box_is_being_defined = False
        while self.source.HasCapture():
            self.source.ReleaseMouse()

    def set_cursor(self, mode=None):
        if (self.forced_cursor is not None):
            self.SetCursor(self.forced_cursor)
            #
            return

        if mode is None:
            mode = self.mouse_mode
        c = mode.get_cursor()
        self.SetCursor(c)

    def get_effective_tool_mode(self, event):
        middle_down = False
        alt_down = False
        if (event is not None):
            try:
                alt_down = event.AltDown()
                # print self.is_alt_key_down
            except:
                pass
            try:
                middle_down = event.MiddleIsDown()
            except:
                pass
        if alt_down or middle_down:
            mode = self.default_mouse_mode
        else:
            mode = self.mouse_mode
        return mode

    def on_left_down(self, evt):
        self.capture_mouse()
        if self.is_editing_in_cell and self.mouse_event_in_edit_cell(evt):
            self.is_mousing_while_editing = True
            self.on_left_down_in_edit_cell(evt)
        else:
            self.end_editing()
            mode = self.get_effective_tool_mode(evt)
            log.debug("on_left_down: effective mode=%s" % mode)
            self.forced_cursor = None
            self.selection_box_is_being_defined = False
            self.mouse_down_position = evt.GetPosition()
            self.mouse_move_position = self.mouse_down_position

            mode.process_left_down(evt)
            self.set_cursor(mode)
        evt.Skip()

    def on_motion(self, evt):
        if self.is_mousing_while_editing:
            self.on_motion_in_edit_cell(evt)
        else:
            mode = self.get_effective_tool_mode(evt)
            # log.debug("on_motion: effective mode=%s" % mode)
            if evt.LeftIsDown() and self.has_capture():
                mode.process_mouse_motion_down(evt)
            else:
                mode.process_mouse_motion_up(evt)
            self.set_cursor(mode)
        evt.Skip()

    def on_left_up(self, evt):
        self.stop_scroll_timer()
        if not self.has_capture():
            return
        self.release_mouse()
        if self.is_mousing_while_editing:
            self.is_mousing_while_editing = False
            self.on_left_up_in_edit_cell(evt)
        else:
            mode = self.get_effective_tool_mode(evt)
            log.debug("on_left_up: effective mode=%s" % mode)
            self.forced_cursor = None
            mode.process_left_up(evt)
            self.set_cursor(mode)
        evt.Skip()

    def on_left_dclick(self, evt):
        # self.SetFocus() # why would it not be focused?
        self.end_editing()
        mode = self.get_effective_tool_mode(evt)
        log.debug("on_left_dclick: effective mode=%s" % mode)
        mode.process_left_dclick(evt)
        self.set_cursor(mode)
        evt.Skip()

    def on_popup(self, evt):
        self.end_editing()
        mode = self.get_effective_tool_mode(evt)
        log.debug("on_popup: effective mode=%s" % mode)
        self.forced_cursor = None
        mode.process_popup(evt)
        self.set_cursor(mode)

    def on_mouse_wheel(self, evt):
        mode = self.get_effective_tool_mode(evt)
        log.debug("on_mouse_wheel: effective mode=%s" % mode)
        mode.process_mouse_wheel(evt)
        self.set_cursor(mode)

    def on_mouse_enter(self, evt):
        self.set_cursor()
        evt.Skip()

    def on_mouse_leave(self, evt):
        self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))
        self.mouse_mode.process_mouse_leave(evt)
        evt.Skip()

    def on_focus(self, evt):
        mode = self.get_effective_tool_mode(evt)
        mode.process_focus(evt)

    def on_focus_lost(self, evt):
        mode = self.get_effective_tool_mode(evt)
        mode.process_focus_lost(evt)

    ##### mouse processing while in cell edit mode

    def mouse_event_in_edit_cell(self, evt):
        return False

    def on_left_down_in_edit_cell(self, evt):
        pass

    def on_motion_in_edit_cell(self, evt):
        pass

    def on_left_up_in_edit_cell(self, evt):
        pass

    ##### autoscrolling

    def stop_scroll_timer(self):
        pass

    ##### command processor

    def create_mouse_event_flags(self):
        return self.segment_viewer.create_mouse_event_flags()

    def handle_motion_update_status(self, evt, row, col):
        msg = self.get_status_message_at_cell(row, col)
        if msg:
            self.segment_viewer.show_status_message(msg)

    ##### Default selection handlers

    def handle_select_start(self, evt, row, col, flags=None):
        """ select_handler: interface with set_style_ranges to highlight bytes
        that should be selected (e.g. rect select will need different method
        than regular selection).

        caret_handler: object that implements the CaretHandler API
        """
        if flags is None:
            flags = self.create_mouse_event_flags()
        ch = self.caret_handler
        try:
            r, c, index1, index2, inside = self.get_location_from_event(evt)
        except IndexError:
            log.warning("mouse event on invalid index")
            return
        self.mouse_drag_started = True
        if c < 0 or flags.selecting_rows or not inside:
            c = 0
            selecting_rows = True
        else:
            selecting_rows = False
        log.debug("handle_select_start: rows=%s, input=%d,%d r,c=%d,%d index1,2=%d,%d inside=%s" % (selecting_rows, row, col, r, c, index1, index2, inside))
        if evt.ControlDown():
            self.multi_select_mode = True
            self.select_extend_mode = False
        elif evt.ShiftDown():
            self.multi_select_mode = False
            self.select_extend_mode = True
        log.debug(("start before:", ch.carets, "multi", self.multi_select_mode, "extend", self.select_extend_mode))
        if self.select_extend_mode:
            caret = self.caret_with_selection = ch.carets.current
            if index1 < caret.anchor_start_index:
                self.select_extend_mode = "bottom anchor"
                if not caret.has_selection:
                    caret.anchor_end_index += 1
                caret.anchor_start_index = index1
                caret.index = index1
            else:
                self.select_extend_mode = "top anchor"
                caret.anchor_end_index = index2
                caret.index = index2 - 1
            caret.anchor_initial_start_index, caret.anchor_initial_end_index = caret.anchor_start_index, caret.anchor_end_index
        else:
            if selecting_rows:
                index1, index2 = self.get_start_end_index_of_row(r)
                caret = Caret(index2)
                ch.carets.force_single_caret(caret)
                self.caret_with_selection = caret
                caret.anchor_initial_start_index, caret.anchor_initial_end_index = index1, index2
                caret.anchor_start_index, caret.anchor_end_index = index1, index2
                flags.keep_selection = True
                log.debug("handle_select_start selecting rows: flags: %s, anchors=%s" % (flags, str((caret.anchor_initial_start_index, caret.anchor_initial_end_index))))
            else:
                flags.keep_selection = True
                # initial click when not selecting rows should move the caret,
                # not select the grid square
                caret = Caret(index1)
                if self.multi_select_mode:
                    ch.add_caret(caret)
                    log.debug(("adding caret", caret))
                else:
                    ch.force_single_caret(caret)
                    log.debug(("forced single caret", caret))
                self.caret_with_selection = caret
                self.pending_select_awaiting_drag = (index1, index2)
                log.debug("handle_select_start placing cursor: flags: %s, index=%s" % (flags, index1))
        flags.caret_column = c
        self.commit_change(flags)
        log.debug(f"start after: {ch.carets}")

    def handle_select_motion(self, evt, row, col, flags=None):
        if not self.mouse_drag_started:
            # On windows, it's possible to get a motion event before a mouse
            # down event, so need this flag to check
            return
        if flags is None:
            flags = self.create_mouse_event_flags()
        ch = self.caret_handler
        update = False
        try:
            r, c, index1, index2, inside = self.get_location_from_col(row, col)
        except IndexError:
            log.debug("mouse location is a hidden cell; skipping")
            return
        log.debug("handle_select_motion: r=%d c=%d index1: %s, index2: %s pending: %s, sel rows: %s" % (r, c, index1, index2, str(self.pending_select_awaiting_drag), flags.selecting_rows))
        # log.debug("handle_select_motion: r=%d c=%d index1: %s, index2: %s pending: %s, sel rows: %s anchors: initial=%s current=%s" % (r, c, index1, index2, str(self.pending_select_awaiting_drag), flags.selecting_rows, str((caret.anchor_initial_start_index, caret.anchor_initial_end_index)), str((caret.anchor_start_index, caret.anchor_end_index))))
        log.debug(("motion before:", ch.carets))
        caret = self.caret_with_selection
        if c < 0 or flags.selecting_rows or not inside:
            selecting_rows = True
            index1, index2 = self.get_start_end_index_of_row(r)
            c = 0
        else:
            selecting_rows = False
            if self.pending_select_awaiting_drag is not None:
                # only start selection if the cursor is over a different cell
                # than the mouse down event
                if index1 == self.pending_select_awaiting_drag[0]:
                    return

                # We have an actual drag so we can begin the selection
                caret.anchor_initial_start_index, caret.anchor_initial_end_index = self.pending_select_awaiting_drag
                caret.anchor_start_index, caret.anchor_end_index = self.pending_select_awaiting_drag
                self.pending_select_awaiting_drag = None
                update = True

        if self.select_extend_mode == "top anchor":
            if index1 < caret.anchor_initial_start_index:
                caret.index = index1
                caret.anchor_start_index = index1
                caret.anchor_end_index = caret.anchor_initial_start_index + 1
                update = True
            else:
                caret.index = index2 - 1
                caret.anchor_start_index = caret.anchor_initial_start_index
                caret.anchor_end_index = index2
                update = True
        elif self.select_extend_mode == "bottom anchor":
            if index1 < caret.anchor_initial_end_index:
                caret.index = index1
                caret.anchor_start_index = index1
                caret.anchor_end_index = caret.anchor_initial_end_index
                update = True
            else:
                caret.index = index2 - 1
                caret.anchor_start_index = caret.anchor_initial_end_index
                caret.anchor_end_index = index2
                update = True
        else:
            if index2 >= caret.anchor_initial_end_index:
                caret.index = index2 - 1
                if index2 != caret.anchor_end_index:
                    caret.anchor_start_index = caret.anchor_initial_start_index
                    caret.anchor_end_index = index2
                    update = True
            elif index1 <= caret.anchor_initial_start_index:
                caret.index = index1
                if index1 != caret.anchor_start_index:
                    caret.anchor_start_index = index1
                    caret.anchor_end_index = caret.anchor_initial_end_index
                    update = True

        if update:
            self.commit_change(flags)
        log.debug("handle_select_motion: update: %s, flags: %s, anchors: initial=%s current=%s" % (update, flags, str((caret.anchor_initial_start_index, caret.anchor_initial_end_index)), str((caret.anchor_start_index, caret.anchor_end_index))))
        # log.debug(f"motion after: {ch.carets}")

    def handle_select_end(self, evt, row, col, flags=None):
        if flags is None:
            flags = self.create_mouse_event_flags()
        self.mouse_drag_started = False
        self.select_extend_mode = False
        self.multi_select_mode = False
        self.caret_with_selection = None
        self.caret_handler.carets.collapse_overlapping()
        if wx.Platform == "__WXMSW__":
            # FIXME: MSW doesn't seem to refresh after a mouse release
            # outside of the window, so force it here to fill in the remaining
            # bits of the selection
            log.debug("Extra refresh on handle_select_end for windows")
            self.refresh_view()
        self.commit_change(flags)
        log.debug(("end after:", self.caret_handler.carets))

    def commit_change(self, flags):
        log.debug(("commit before:", self.caret_handler.carets))
        self.refresh_ranges(self.caret_handler)
        self.caret_handler.sync_caret_event = flags
        self.caret_handler.ensure_visible_event = flags
        self.caret_handler.refresh_event = flags
        log.debug(("commit after:", self.caret_handler.carets))

    def get_location_from_event(self, evt):
        raise NotImplementedError

    def get_start_end_index_of_row(self, row):
        raise NotImplementedError

    def get_status_message_at_cell(self, row, col):
        raise NotImplementedError

    def highlight_selected_ranges(self, caret_handler):
        ch = self.caret_handler
        ch.document.change_count += 1
        s = ch.segment
        s.clear_style_bits(selected=True)
        self.segment_viewer.highlight_selected_ranges_in_segment(ch.carets.selected_ranges, s)
        ch.calc_dependent_action_enabled_flags()



# Warning! This is generated code.
# 
# Any edits will be overwritten with the next call to parse_gen.py


cpu_id_to_name = {10: '6502', 11: '6502 (including undocumented instructions)', 12: '65816', 13: '65C02', 14: '6800', 15: '6809', 16: '6811', 17: '8051', 18: '8080', 32: 'Jumpman Level', 0: 'Data', 30: 'Antic Display List', 31: 'Jumpman Harvest Table'}

cpu_name_to_id = {'6502': 10, '6502 (including undocumented instructions)': 11, '65816': 12, '65C02': 13, '6800': 14, '6809': 15, '6811': 16, '8051': 17, '8080': 18, 'Jumpman Level': 32, 'Data': 0, 'Antic Display List': 30, 'Jumpman Harvest Table': 31}

valid_cpu_ids = [10, 11, 12, 13, 14, 15, 16, 17, 18]



import numpy as np

import logging
log = logging.getLogger(__name__)


user_bit_mask = 0x07
not_user_bit_mask = 0xff ^ user_bit_mask
data_bit_mask = 0x08
diff_bit_mask = 0x10
match_bit_mask = 0x20
comment_bit_mask = 0x40
selected_bit_mask = 0x80


def get_style_bits(match=False, comment=False, selected=False, data=False, diff=False, user=0):
    """ Return an int value that contains the specified style bits set.

    Available styles for each byte are:

    match: part of the currently matched search
    comment: user commented area
    selected: selected region
    data: labeled in the disassembler as a data region (i.e. not disassembled)
    """
    style_bits = 0
    if user:
        style_bits |= (user & user_bit_mask)
    if diff:
        style_bits |= diff_bit_mask
    if match:
        style_bits |= match_bit_mask
    if comment:
        style_bits |= comment_bit_mask
    if data:
        style_bits |= data_bit_mask
    if selected:
        style_bits |= selected_bit_mask
    return style_bits


def get_style_mask(**kwargs):
    """Get the bit mask that, when anded with data, will turn off the
    selected bits
    """
    bits = get_style_bits(**kwargs)
    if 'user' in kwargs and kwargs['user']:
        bits |= user_bit_mask
    else:
        bits &= (0xff ^ user_bit_mask)
    return 0xff ^ bits
